"""Temporary!"""
import os

from flask import request
from lexrpc.base import XrpcError
from lexrpc.server import Server

from .util import parse_at_uri, TombstonedRepo


# XRPC server
server = Server(validate=True)

# initialized in app.py, testutil.XrpcTestCase.setUp
storage = None


def auth():
    token = os.environ.get('REPO_TOKEN')
    if not token:
      raise NotImplementedError(
          'Authenticated XRPC methods are not currently supported')

    if request.headers.get('Authorization') != f'Bearer {token}':
        raise ValueError('Invalid bearer token in Authorization header')


def load_repo(did_or_at_uri):
    if did_or_at_uri.startswith('at://'):
        did_or_handle, _, _ = parse_at_uri(did_or_at_uri)
    else:
        did_or_handle = did_or_at_uri

    try:
        repo = storage.load_repo(did_or_handle)
    except TombstonedRepo:
        raise XrpcError(f'Repo {did_or_handle} is tombstoned', name='RepoDeactivated')

    if not repo:
        raise XrpcError(f'Repo {did_or_handle} not found', name='RepoNotFound')

    return repo



"""Bluesky repo storage base class and in-memory implementation.

Lightly based on:
https://github.com/bluesky-social/atproto/blob/main/packages/repo/src/storage/repo-storage.ts
"""
from collections import namedtuple
from enum import auto, Enum
import itertools

import dag_cbor
from multiformats import CID, multicodec, multihash

from . import util
from .util import dag_cbor_cid, DEACTIVATED, tid_to_int, TOMBSTONED, TombstonedRepo

SUBSCRIBE_REPOS_NSID = 'com.atproto.sync.subscribeRepos'


class Action(Enum):
    """Used in :meth:`Repo.format_commit`.

    TODO: switch to StrEnum once we can require Python 3.11.
    """
    CREATE = auto()
    UPDATE = auto()
    DELETE = auto()

# TODO: Should this be a subclass of Block?
# TODO: generalize to handle other events
CommitData = namedtuple('CommitData', [
    'commit',  # Block
    'blocks',  # dict of CID to Block
    'prev',    # CID or None
], defaults=[None])  # for ops

CommitOp = namedtuple('CommitOp', [  # for subscribeRepos
    'action',  # Action
    'path',    # str
    'cid',     # CID, or None for DELETE
])

# commit record format is:
# https://atproto.com/specs/repository#commit-objects
#
# {
#     'version': 3,
#     'did': [repo],
#     'rev': [str, TID],
#     'data': [CID],
#     'prev': [CID or None],
#     'sig': [bytes],
# }


class Block:
    r"""An ATProto block: a record, :class:`MST` entry, commit, or other event.

    Can start from either encoded bytes or decoded object, with or without
    :class:`CID`. Decodes, encodes, and generates :class:`CID` lazily, on
    demand, on attribute access.

    Events should have a fully-qualified ``$type`` field that's one of the
    ``message`` types in ``com.atproto.sync.subscribeRepos``, eg
    ``com.atproto.sync.subscribeRepos#tombstone``.

    Based on :class:`carbox.car.Block`.

    Attributes:
      cid (CID): lazy-loaded (dynamic property)
      decoded (dict): decoded object (dynamic property)
      encoded (bytes): DAG-CBOR encoded data (dynamic property)
      seq (int): ``com.atproto.sync.subscribeRepos`` sequence number
      ops (list): :class:`CommitOp`\s if this is a commit, otherwise None
      time (datetime): when this block was first created
      repo (str): DID of a repo that includes this block. Occasionally, blocks
        may be included in more than one repo, so this may be *any* repo that
        includes it. In practice, it's often the first or last repo that
        included it.
    """
    def __init__(self, *, cid=None, decoded=None, encoded=None, seq=None,
                 ops=None, time=None, repo=None):
        """Constructor.

        Args:
          cid (CID): optional
          decoded (dict): optional
          encoded (bytes): optional
        """
        assert encoded or decoded
        self._cid = cid
        self._encoded = encoded
        self._decoded = decoded
        self.seq = seq
        self.ops = ops
        self.time = time or util.now()
        self.repo = repo

    def __str__(self):
        return f'<Block: {self.cid}>'

    @property
    def cid(self):
        if self._cid is None:
            digest = multihash.digest(self.encoded, 'sha2-256')
            self._cid = CID('base58btc', 1, 'dag-cbor', digest)
        return self._cid

    @property
    def encoded(self):
        if self._encoded is None:
            self._encoded = dag_cbor.encode(self.decoded)
        return self._encoded

    @property
    def decoded(self):
        if self._decoded is None:
            self._decoded = dag_cbor.decode(self.encoded)
        return self._decoded

    def __eq__(self, other):
        """Compares by CID only."""
        return self.cid == other.cid

    def __hash__(self):
        return hash(self.cid)


class Storage:
    """Abstract base class for storing nodes: records, MST entries, commits, etc.

    Concrete subclasses should implement this on top of physical storage,
    eg database, filesystem, in memory.

    TODO: batch operations?

    Attributes:
      head (CID)
    """
    head = None

    def create_repo(self, repo, *, signing_key, rotation_key=None):
        """Stores a new repo's metadata in storage.

        Only stores the repo's DID, handle, and head commit :class:`CID`, not
        blocks!

        If the repo already exists in storage, this should update it instead of
        failing.

        Args:
          repo (Repo)
          signing_key (ec.EllipticCurvePrivateKey)
          rotation_key (ec.EllipticCurvePrivateKey): optional
        """
        raise NotImplementedError()

    def load_repo(self, did_or_handle):
        """Loads a repo from storage.

        Args:
          did_or_handle (str): optional

        Returns:
          Repo, or None if the did or handle wasn't found:

        Raises:
          TombstonedRepo: if the repo is tombstoned
        """
        raise NotImplementedError()

    def load_repos(self, after=None, limit=500):
        """Loads multiple repos from storage.

        Repos are returned in lexicographic order of their DIDs, ascending.
        Tombstoned repos are included.

        Args:
          after (str): optional DID to start at, *exclusive*
          limit (int): maximum number of repos to return

        Returns:
          sequence of Repo:
        """
        raise NotImplementedError()

    def deactivate_repo(self, repo):
        """Marks a repo as deactivated.

        * Stores a ``com.atproto.sync.subscribeRepos#account`` block with its
          own sequence number.
        * If :attr:`Repo.callback` is populated, calls it with the
          ``com.atproto.sync.subscribeRepos#account`` message.
        * Calls :meth:`Repo._set_status` to mark the repo as deactivated in storage.

        Args:
          repo (Repo)
        """
        self._set_repo_status(repo, DEACTIVATED)
        block = self.write_event(repo=repo, type='account',
                                 active=False, status='deactivated')

    def activate_repo(self, repo):
        """Marks a repo as active.

        Only needed after deactivating. Does nothing if the repo is tombstoned.

        * Stores a ``com.atproto.sync.subscribeRepos#account`` block with its
          own sequence number.
        * If :attr:`Repo.callback` is populated, calls it with the
          ``com.atproto.sync.subscribeRepos#account`` message.
        * Calls :meth:`Repo._set_status` to mark the repo as active in storage.

        Args:
          repo (Repo)
        """
        self._set_repo_status(repo, None)
        block = self.write_event(repo=repo, type='account', active=True)

    def tombstone_repo(self, repo):
        """Marks a repo as tombstoned.

        * Stores a ``com.atproto.sync.subscribeRepos#tombstone`` block with its
          own sequence number.
        * If :attr:`Repo.callback` is populated, calls it with the
          ``com.atproto.sync.subscribeRepos#tombstone`` message.
        * Calls :meth:`Repo._set_status` to mark the repo as deactivated in storage.

        After this, :meth:`load_repo` will raise :class:`TombstonedRepo` for
        this repo.

        Args:
          repo (Repo)
        """
        self._set_repo_status(repo, TOMBSTONED)
        block = self.write_event(repo=repo, type='tombstone')

    def _tombstone_repo(self, repo):
        """Marks a repo as tombstoned in storage.

        Args:
          repo (Repo)
        """
        raise NotImplementedError()

    def read(self, cid):
        """Reads a node from storage.

        Args:
          cid (CID)

        Returns:
          Block, or None if not found:
        """
        raise NotImplementedError()

    def read_many(self, cids, require_all=True):
        """Batch read multiple nodes from storage.

        Args:
          cids (sequence of CID)
          require_all (bool): whether to assert that all cids are found

        Returns:
          dict: {:class:`CID`: :class:`Block` or None if not found}
        """
        raise NotImplementedError()

    def read_blocks_by_seq(self, start=0, repo=None):
        """Batch read blocks from storage by ``subscribeRepos`` sequence number.

        Args:
          seq (int): optional ``subscribeRepos`` sequence number to start from.
            Defaults to 0.
          repo (str): optional repo DID. If not provided, all repos are included.

        Returns:
          iterable or generator: all :class:`Block` s starting from ``seq``,
          inclusive, in ascending ``seq`` order
        """
        raise NotImplementedError()

    def read_events_by_seq(self, start=0, repo=None):
        """Batch read commits and other events by ``subscribeRepos`` sequence number.

        Args:
          start (int): optional ``subscribeRepos`` sequence number to start from,
            inclusive. Defaults to 0.
          repo (str): optional repo DID. If not provided, all repos are included.

        Returns:
          generator: generator of :class:`CommitData` for commits and dict
          messages for other events, starting from ``seq``, inclusive, in
          ascending ``seq`` order
        """
        assert start >= 0

        seq = commit_block = blocks = None

        def make_commit():
            for op in commit_block.ops:
                if (op.action in (Action.CREATE, Action.UPDATE)
                        and op.cid not in blocks):
                    record = self.read(op.cid)
                    assert record
                    blocks[op.cid] = record
            return CommitData(blocks=blocks, commit=commit_block,
                              prev=commit_block.decoded.get('prev'))

        seen = []  # CIDs
        for block in self.read_blocks_by_seq(start=start, repo=repo):
            assert block.seq
            if block.seq != seq:  # switching to a new commit's blocks
                if commit_block:
                    yield make_commit()
                else:
                    # we shouldn't have any dangling blocks that we don't serve
                    assert not blocks
                seq = block.seq
                blocks = {}  # maps CID to Block
                commit_block = None

            if block.decoded.get('$type', '').startswith(
                    'com.atproto.sync.subscribeRepos#'):  # non-commit message
                yield block.decoded
                continue

            blocks[block.cid] = block
            commit_fields = ['version', 'did', 'rev', 'prev', 'data', 'sig']
            if block.decoded.keys() == set(commit_fields):
                commit_block = block

        # final commit
        if blocks:
            assert commit_block, f'seq {seq}'
            yield make_commit()

    def has(self, cid):
        """Checks if a given :class:`CID` is currently stored.

        Args:
          cid (CID)

        Returns:
          bool:
        """
        raise NotImplementedError()

    def write(self, repo_did, obj, seq=None):
        """Writes a node to storage.

        Args:
          repo_did (str):
          obj (dict): a record, commit, serialized :class:`MST` node, or
            `subscribeRepos` event/message
          seq (int or None): sequence number. If not provided, a new one will be
            allocated.

        Returns:
          Block:
        """
        raise NotImplementedError()

    def write_event(self, repo, type, **kwargs):
        """Writes a ``subscribeRepos`` event to storage.

        Args:
          repo (Repo)
          type (str): ``account`` or ``identity``
          kwargs: included in the event, eg ``active``, `status``

        Returns:
          Block:
        """
        assert type in ('account', 'identity', 'tombstone'), type

        seq = self.allocate_seq(SUBSCRIBE_REPOS_NSID)
        block = self.write(repo.did, {
            '$type': f'com.atproto.sync.subscribeRepos#{type}',
            'seq': seq,
            'did': repo.did,
            'time': util.now().isoformat(),
            **kwargs,
        }, seq=seq)

        if repo.callback:
            repo.callback(block.decoded)
        return block

    def apply_commit(self, commit_data):
        """Writes a commit to storage.

        Generates a new sequence number and uses it for all blocks in the commit.

        Args:
          commit (CommitData)
        """
        raise NotImplementedError()

    def allocate_seq(self, nsid):
        """Generates and returns a sequence number for the given NSID.

        Sequence numbers must be monotonically increasing positive integers, per
        NSID. They may have gaps. Background:
        https://atproto.com/specs/event-stream#sequence-numbers

        Args:
          nsid (str): subscription XRPC method this sequence number is for

        Returns:
          int:
        """
        raise NotImplementedError()

    def last_seq(self, nsid):
        """Returns the last (highest) stored sequence number for the given NSID.

        Args:
          nsid (str): subscription XRPC method this sequence number is for

        Returns:
          int:
        """
        raise NotImplementedError()


class MemoryStorage(Storage):
    """In memory storage implementation.

    Attributes:
      repos (list of :class:`Repo`)
      blocks (dict): {:class:`CID`: :class:`Block`}
      head (CID)
      sequences (dict): {str NSID: int next sequence number}
    """
    repos = None
    blocks = None
    head = None
    sequences = None

    def __init__(self):
        self.blocks = {}
        self.repos = []
        self.sequences = {}

    def create_repo(self, repo, *, signing_key, rotation_key=None):
        if repo not in self.repos:
            self.repos.append(repo)

    def load_repo(self, did_or_handle):
        assert did_or_handle

        for repo in self.repos:
            if did_or_handle in (repo.did, repo.handle):
                if repo.status == TOMBSTONED:
                    raise TombstonedRepo(f'{repo.did} is tombstoned')
                return repo

    def load_repos(self, after=None, limit=500):
        it = iter(sorted(self.repos, key=lambda repo: repo.did))

        if after:
            it = itertools.dropwhile(lambda repo: repo.did <= after, it)

        return list(itertools.islice(it, limit))

    def _set_repo_status(self, repo, status):
        repo.status = status

    def read(self, cid):
        return self.blocks.get(cid)

    def read_many(self, cids, require_all=True):
        cids = list(cids)
        found = {cid: self.blocks.get(cid) for cid in cids}
        if require_all:
            assert len(found) == len(cids), (len(found), len(cids))
        return found

    def read_blocks_by_seq(self, start=0, repo=None):
        assert start >= 0
        return sorted((b for b in self.blocks.values()
                       if b.seq >= start and (not repo or b.repo == repo)),
                      key=lambda b: b.seq)

    def has(self, cid):
        return cid in self.blocks

    def write(self, repo_did, obj, seq=None):
        if seq is None:
            seq = self.allocate_seq(SUBSCRIBE_REPOS_NSID)

        block = Block(decoded=obj, seq=seq, repo=repo_did)
        if block not in self.blocks:
            self.blocks[block.cid] = block
        return block

    def apply_commit(self, commit_data):
        seq = tid_to_int(commit_data.commit.decoded['rev'])
        assert seq

        for block in commit_data.blocks.values():
            block.seq = seq

        # only add new blocks so we don't wipe out any existing blocks' sequence
        # numbers. (occasionally we see existing blocks recur, eg MST nodes.)
        for cid, block in commit_data.blocks.items():
            self.blocks.setdefault(cid, block)

        self.head = commit_data.commit.cid
        # the Repo will generally already be in self.repos, and it updates its
        # own head cid, so no need to do that here manually.

    def allocate_seq(self, nsid):
        assert nsid
        next = self.sequences.setdefault(nsid, 1)
        self.sequences[nsid] += 1
        return next

    def last_seq(self, nsid):
        assert nsid
        return self.sequences[nsid] - 1



"""Bluesky / AT Protocol Merkle search tree implementation.

* https://atproto.com/guides/data-repos
* https://atproto.com/lexicons/com-atproto-sync
* https://hal.inria.fr/hal-02303490/document

Heavily based on:
https://github.com/bluesky-social/atproto/blob/main/packages/repo/src/mst/mst.ts

Huge thanks to the Bluesky team for working in the public, in open source, and to
Daniel Holmgren and Devin Ivy for this code specifically!

From that file:

This is an implementation of a Merkle Search Tree (MST)
The data structure is described here: https://hal.inria.fr/hal-02303490/document
The MST is an ordered, insert-order-independent, deterministic tree.
Data keys are laid out in alphabetic order.
The key insight of an MST is that each key is hashed and starting 0s are counted
to determine which layer it falls on (5 zeros for ~32 fanout).
This is a merkle tree, so each subtree is referred to by it's hash (CID).
When a leaf is changed, ever tree on the path to that leaf is changed as well,
thereby updating the root hash.

For atproto, we use SHA-256 as the key hashing algorithm, and ~4 fanout
(2-bits of zero per layer).

A couple notes on CBOR encoding:

There are never two neighboring subtrees.
Therefore, we can represent a node as an array of
leaves & pointers to their right neighbor (possibly null),
along with a pointer to the left-most subtree (also possibly null).

Most keys in a subtree will have overlap.
We do compression on prefixes by describing keys as:
* the length of the prefix that it shares in common with the preceding key
* the rest of the string

For example:

If the first leaf in a tree is ``bsky/posts/abcdefg`` and the second is
``bsky/posts/abcdehi``, then the first will be described as ``prefix: 0, key:
'bsky/posts/abcdefg'``, and the second will be described as ``prefix: 16, key:
'hi'``.
"""
from collections import namedtuple
import copy
from hashlib import sha256
import logging
from os.path import commonprefix
import re

import dag_cbor
from multiformats import CID

from .storage import Block, Storage
from .util import dag_cbor_cid

logger = logging.getLogger(__name__)

# this is treeEntry in mst.ts
Entry = namedtuple('Entry', [
    'p',  # int, length of prefix that this data key shares with the prev data key
    'k',  # bytes, the rest of the data key outside the shared prefix
    'v',  # str CID, value
    't',  # str CID, next subtree (to the right of leaf), or None
])

Data = namedtuple('Data', [
    'l',  # str CID, left-most subtree, or None
    'e',  # list of Entry
])

Leaf = namedtuple('Leaf', [
    'key',    # str, data key (collection + record key aka rkey)
    'value',  # CID
])


class MST:
    """Merkle search tree class.

    Attributes:
      storage (Storage):
      entries (sequence of MST and Leaf)
      layer (int): this MST's layer in the root MST
      pointer (CID):
      outdated_pointer (bool): whether pointer needs to be recalculated
    """
    storage = None
    entries = None
    layer = None
    pointer = None
    outdated_pointer = False

    def __init__(self, *, storage=None, entries=None, pointer=None, layer=None):
        """Constructor.

        Args:
          storage (Storage)
          entries (sequence of MST and Leaf)
          pointer (CID)
          layer (int)

        Returns:
          MST:
        """
        self.storage = storage
        self.entries = entries
        self.pointer = pointer
        self.layer = layer

    @classmethod
    def load(cls, *, storage=None, cid=None):
        return MST(storage=storage, entries=None, pointer=cid, layer=None)

    @classmethod
    def create(cls, *, storage=None, entries=None, layer=None):
        """

        Args:
          storage (Storage)
          entries (sequence of MST and Leaf)
          layer (int)

        Returns:
          MST
        """
        if not entries:
            entries = []
        pointer = cid_for_entries(entries)
        return MST(storage=storage, entries=entries, pointer=pointer, layer=layer)

#     def from_data(storage, data, opts):
#         """
#         Returns:
#           MST:
#         """
#         entries = deserialize_node_data(data)
#         pointer = cid_for_cbor(data)
#         return MST(entries=entries, pointer=pointer)

    def __eq__(self, other):
        if isinstance(other, MST):
            return self.get_pointer() == other.get_pointer()

    def __unicode__(self):
        return f'MST with pointer {self.get_pointer()}'

    def __repr__(self):
        return f'MST(storage={self.storage}, entries=..., pointer={self.get_pointer()}, layer={self.get_layer()})'

    # Immutability
    # -------------------
    def new_tree(self, entries):
        """We never mutate an MST, we just return a new MST with updated values.

        Args:
            entries (sequence of MST and Leaf)

        Returns:
            MST:
        """
        mst = MST(storage=self.storage, entries=entries, pointer=self.pointer,
                  layer=self.layer)
        mst.outdated_pointer = True
        return mst


#     Getters (lazy load)
#     -------------------

    def get_entries(self):
        """

        We don't want to load entries of every subtree, just the ones we need.

        Returns:
          sequence of MST and Leaf:
        """
        if self.entries is not None:
            return copy.copy(self.entries)

        if self.pointer:
            data = Data(**self.storage.read(self.pointer).decoded)
            first_leaf = layer = None
            if data.e:
                layer = leading_zeros_on_hash(data.e[0]['k'])

            self.entries = deserialize_node_data(storage=self.storage, data=data,
                                                 layer=layer)
            return self.entries

        raise RuntimeError('No entries or CID provided')

    def get_pointer(self):
        """Returns this MST's root CID pointer. Calculates it if necessary.

        We don't hash the node on every mutation for performance reasons.
        Instead we keep track of whether the pointer is outdated and only
        (recursively) calculate when needed.

        Returns:
          CID:
        """
        if not self.outdated_pointer:
            return self.pointer

        outdated = False
        entries = self.get_entries()
        for e in entries:
            if isinstance(e, MST) and e.outdated_pointer:
                outdated = True
                e.get_pointer()

        if outdated:
            entries = self.get_entries()

        self.pointer = cid_for_entries(entries)
        self.outdated_pointer = False
        return self.pointer

    def get_layer(self):
        """Returns this MST's layer, and sets ``self.layer``.

        In most cases, we get the layer of a node from a hint on creation. In the
        case of the topmost node in the tree, we look for a key in the node &
        determine the layer. In the case where we don't find one, we recurse down
        until we do. If we still can't find one, then we have an empty tree and the
        node is layer 0.

        Returns:
          int:
        """
        self.layer = self.attempt_get_layer()
        if self.layer is None:
            self.layer = 0

        return self.layer

    def attempt_get_layer(self):
        """Returns this MST's layer, and sets ``self.layer``.

        Returns:
          int or None:
        """
        if self.layer is not None:
            return self.layer

        entries = self.get_entries()
        layer = layer_for_entries(entries)
        if layer is None:
            for entry in entries:
                if isinstance(entry, MST):
                    child_layer = entry.attempt_get_layer()
                    if child_layer is not None:
                        layer = child_layer + 1
                        break

        if layer is not None:
            self.layer = layer

        return layer


    # Core functionality
    # -------------------

    def get_unstored_blocks(self):
        """Return the necessary blocks to persist the MST to repo storage.

        Returns:
          (CID root, dict mapping CID to Block) tuple:
        """
        unstored = {}
        pointer = self.get_pointer()

        if self.storage.has(pointer):
            return pointer, unstored

        entries = self.get_entries()
        data = serialize_node_data(entries)
        block = Block(decoded=data._asdict())
        unstored[block.cid] = block

        for entry in entries:
            if isinstance(entry, MST):
                _, blocks = entry.get_unstored_blocks()
                unstored.update(blocks)

        return pointer, unstored

    def add(self, key, value=None, known_zeros=None):
        """Adds a new leaf for the given key/value pair.

        Args:
          key (str)
          value (CID)
          known_zeros (int)

        Returns:
          MST:

        Raises:
          ValueError: if a leaf with that key already exists
        """
        ensure_valid_key(key)
        key_zeros = known_zeros or leading_zeros_on_hash(key)
        layer = self.get_layer()
        new_leaf = Leaf(key=key, value=value)

        if key_zeros == layer:
            # it belongs in self layer
            index = self.find_gt_or_equal_leaf_index(key)
            found = self.at_index(index)
            if isinstance(found, Leaf) and found.key == key:
                raise ValueError(f'There is already a value at key: {key}')
            prev_node = self.at_index(index - 1)
            if not prev_node or isinstance(prev_node, Leaf):
                # if entry before is a leaf, (or we're on far left) we can just splice in
                return self.splice_in(new_leaf, index)
            else:
                # else we try to split the subtree around the key
                left, right = prev_node.split_around(key)
                return self.replace_with_split(index - 1, left, new_leaf, right)

        elif key_zeros < layer:
            # it belongs on a lower layer
            index = self.find_gt_or_equal_leaf_index(key)
            prev_node = self.at_index(index - 1)
            if prev_node and isinstance(prev_node, MST):
                # if entry before is a tree, we add it to that tree
                new_subtree = prev_node.add(key, value, key_zeros)
                return self.update_entry(index - 1, new_subtree)
            else:
                sub_tree = self.create_child()
                new_subtree = sub_tree.add(key, value, key_zeros)
                return self.splice_in(new_subtree, index)

        else:  # key_zeros > layer
            # it belongs on a higher layer, push the rest of the tree down
            left, right = self.split_around(key)
            # if the newly added key has >=2 more leading zeros than the current
            # highest layer then we need to add structural nodes between as well
            layer = self.get_layer()
            extra_layers_to_add = key_zeros - layer
            # intentionally starting at 1, first layer is taken care of by split
            for i in range(1, extra_layers_to_add):
                if left:
                    left = left.create_parent()
                if right:
                    right = right.create_parent()

            updated = []
            if left:
                updated.append(left)
            updated.append(Leaf(key=key, value=value))
            if right:
                updated.append(right)

            new_root = MST.create(storage=self.storage, entries=updated, layer=key_zeros)
            new_root.outdated_pointer = True
            return new_root

    def get(self, key):
        """Gets the value at the given key.

        Args:
          key (str)

        Returns:
          CID or None:
        """
        index = self.find_gt_or_equal_leaf_index(key)
        found = self.at_index(index)
        if found and isinstance(found, Leaf) and found.key == key:
            return found.value

        prev = self.at_index(index - 1)
        if prev and isinstance(prev, MST):
            return prev.get(key)

    def update(self, key, value):
        """Edits the value at the given key.

        Args:
          key (str)
          value (CID)

        Returns:
          MST:

        Raises:
          KeyError: if key doesn't exist
        """
        ensure_valid_key(key)

        index = self.find_gt_or_equal_leaf_index(key)
        found = self.at_index(index)
        if found and isinstance(found, Leaf) and found.key == key:
            return self.update_entry(index, Leaf(key=key, value=value))

        prev = self.at_index(index - 1)
        if prev and isinstance(prev, MST):
            updated_tree = prev.update(key, value)
            return self.update_entry(index - 1, updated_tree)

        raise KeyError(f'Could not find a record with key: {key}')

    def delete(self, key):
        """Deletes the value at the given key.

        Args:
          key (str)

        Returns:
          MST

        Raises:
          KeyError: if key doesn't exist
        """
        return self.delete_recurse(key).trim_top()

    def delete_recurse(self, key):
        """Deletes the value and subtree, if any, at the given key.

        Args:
          key (str):

        Returns:
          MST
        """
        index = self.find_gt_or_equal_leaf_index(key)
        found = self.at_index(index)

        # if found, remove it on self level
        if isinstance(found, Leaf) and found.key == key:
            prev = self.at_index(index - 1)
            next = self.at_index(index + 1)
            if isinstance(prev, MST) and isinstance(next, MST):
                merged = prev.append_merge(next)
                return self.new_tree(
                    self.slice(0, index - 1) + [merged] + self.slice(index + 2)
                )
            else:
                return self.remove_entry(index)

        # else recurse down to find it
        prev = self.at_index(index - 1)
        if isinstance(prev, MST):
            subtree = prev.delete_recurse(key)
            if subtree.get_entries():
                return self.update_entry(index - 1, subtree)
            else:
                return self.remove_entry(index - 1)

        raise KeyError(f'Could not find a record with key: {key}')


#     Simple Operations
#     -------------------

    def update_entry(self, index, entry):
        """Updates an entry in place.

        Args:
          index (int)
          entry (MST or Leaf)

        Returns:
          MST:
        """
        return self.new_tree(
            entries=self.slice(0, index) + [entry] + self.slice(index + 1))

    def remove_entry(self, index):
        """Removes the entry at a given index.

        Args:
          index (int)

        Returns:
          MST:
        """
        return self.new_tree(entries=self.slice(0, index) + self.slice(index + 1))

    def append(self, entry):
        """Appends an entry to the end of the node.

        Args:
          entry (MST or Leaf)

        Returns:
          MST:
        """
        return self.new_tree(self.get_entries() + [entry])

    def prepend(self, entry):
        """Prepends an entry to the start of the node.

        Args:
          entry (MST or Leaf)

        Returns:
          MST:
        """
        return self.new_tree([entry] + self.get_entries())

    def at_index(self, index):
        """Returns the entry at a given index.

        Args:
          index (int)

        Returns:
          MST or Leaf or None:
        """
        entries = self.get_entries()
        if 0 <= index < len(entries):
            return entries[index]

    def slice(self, start=None, end=None):
        """Returns a slice of this node.

        Args:
          start (int): optional, inclusive
          end (int): optional, exclusive

        Returns:
          sequence of MST and Leaf:
        """
        return self.get_entries()[start:end]

    def splice_in(self, entry, index):
        """Inserts an entry at a given index.

        Args:
          entry (MST or Leaf)
          index (int)

        Returns:
          MST:
        """
        return self.new_tree(self.slice(0, index) + [entry] + self.slice(index))

    def replace_with_split(self, index, left=None, leaf=None, right=None):
        """Replaces an entry with [ Maybe(tree), Leaf, Maybe(tree) ].

        Args:
          index (int):
          left (MST or Leaf):
          leaf (Leaf):
          right (MST or Leaf):

        Returns:
          MST:
        """
        updated = self.slice(0, index)
        if left:
            updated.append(left)
        updated.append(leaf)
        if right:
            updated.append(right)
        updated.extend(self.slice(index + 1))
        return self.new_tree(updated)

    def trim_top(self):
        """Trims the top and return its subtree, if necessary.

        Only if the topmost node in the tree only points to another tree.
        Otherwise, does nothing.

        Returns:
          MST:
        """
        entries = self.get_entries()
        if len(entries) == 1 and isinstance(entries[0], MST):
            return entries[0].trim_top()
        else:
            return self


#     Subtree & Splits
#     -------------------

    def split_around(self, key):
        """Recursively splits a subtree around a given key.

        Args:
          key (str)

        Returns:
          (MST or None, MST or None) tuple:
        """
        index = self.find_gt_or_equal_leaf_index(key)
        # split tree around key
        left_data = self.slice(0, index)
        right_data = self.slice(index)
        left = self.new_tree(left_data)
        right = self.new_tree(right_data)

        # if the far right of the left side is a subtree,
        # we need to split it on the key as well
        last_in_left = left_data[-1] if left_data else None
        if isinstance(last_in_left, MST):
            left = left.remove_entry(len(left_data) -1)
            split = last_in_left.split_around(key)
            if split[0]:
                left = left.append(split[0])
            if split[1]:
                right = right.prepend(split[1])

        return [
            left if left.get_entries() else None,
            right if right.get_entries() else None,
        ]

    def append_merge(self, to_merge):
        """Merges another tree with this one.

        The simple merge case where every key in the right tree is greater than
        every key in the left tree. Used primarily for deletes.

        Args:
          to_merge (MST)

        Returns:
          MST:
        """
        assert self.get_layer() == to_merge.get_layer(), \
            'Trying to merge two nodes from different layers of the MST'

        self_entries = self.get_entries()
        to_merge_entries = to_merge.get_entries()
        last_in_left = self_entries[-1]
        first_in_right = to_merge_entries[0]

        if isinstance(last_in_left, MST) and isinstance(first_in_right, MST):
            merged = last_in_left.append_merge(first_in_right)
            return self.new_tree(
                list(self_entries[:-1]) + [merged] + to_merge_entries[1:])
        else:
            return self.new_tree(self_entries + to_merge_entries)


    # Create relatives
    # -------------------

    def create_child(self):
        """
        Returns:
          MST:
        """
        return MST.create(storage=self.storage, entries=[],
                          layer=self.get_layer() - 1)

    def create_parent(self):
        """
        Returns:
          MST:
        """
        parent = MST.create(storage=self.storage, entries=[self],
                            layer=self.get_layer() + 1)
        parent.outdated_pointer = True
        return parent


#     Finding insertion points
#     -------------------

    def find_gt_or_equal_leaf_index(self, key):
        """Finds the index of the first leaf node greater than or equal to value.

        Args:
          key (str)

        Returns:
          int:
        """
        entries = self.get_entries()
        for i, entry in enumerate(entries):
            if isinstance(entry, Leaf) and entry.key >= key:
                return i

        # if we can't find it, we're on the end
        return len(entries)


#     List operations (partial tree traversal)
#     -------------------

    def walk_leaves_from(self, key):
        """Walk tree starting at key.

        Generator for leaves in the tree, starting at a given key.

        Args:
          key (str):

        Generates:
          Leaf
        """
        index = self.find_gt_or_equal_leaf_index(key)
        entries = self.get_entries()

        if index > 0:
            prev = entries[index - 1]
            if prev and isinstance(prev, MST):
                for e in prev.walk_leaves_from(key):
                    yield e

        for entry in entries[index:]:
            if isinstance(entry, Leaf):
                yield entry
            else:
                for e in entry.walk_leaves_from(key):
                    yield e

    def list(self, after=None, before=None):
        """Returns entries, optionally bounded within a key range.

        Args:
          after (str): key, optional
          before (str): key, optional

        Returns:
          sequence of Leaf:
        """
        vals = []

        for leaf in self.walk_leaves_from(after or ''):
            if leaf.key == after:
                continue
            if before and leaf.key >= before:
                break
            vals.append(leaf)

        return vals

    def list_with_prefix(self, prefix):
        """Returns entries with a given key prefix.

        Args:
          prefix (str): key prefix

        Returns:
          sequence of Leaf
        """
        vals = []

        for leaf in self.walk_leaves_from(prefix):
            if not leaf.key.startswith(prefix):
                break
            vals.append(leaf)

        return vals

#     Full tree traversal
#     -------------------

    def walk(self):
        """Walk full tree, depth first, and emit nodes.

        Returns:
          generator of MST and Leaf:
        """
        yield self

        for entry in self.get_entries():
            if isinstance(entry, MST):
                for e in entry.walk():
                    yield e
            else:
                yield entry

#     Walk full tree & emit nodes, consumer can bail at any point by returning False
#     def paths():
#     """
#     Returns:
#       sequence of MST and Leaf
#     """
#         paths = []
#         for entry in self.get_entries():
#             if isinstance(entry, Leaf):
#                 paths.append([entry])
#             if isinstance(entry, MST):
#                 sub_paths = entry.paths()
#                 paths.extend([entry] + p for p in sub_paths)
#
#         return paths

    def all_nodes(self):
        """Walks the tree and returns all nodes.

        Returns:
          sequence of MST and Leaf:
        """
        return list(self.walk())

#     Walks tree & returns all cids
#     def all_cids():
#     """
#     Returns:
#       CidSet
#     """
#         cids = CidSet()
#         for entry in self.get_entries():
#             if isinstance(entry, Leaf):
#                 cids.add(entry.value)
#             else:
#                 subtree_cids = entry.all_cids()
#                 cids.add_set(subtree_cids)
#         cids.add(self.get_pointer())
#         return cids

    def leaves(self):
        """Walks tree and returns all leaves.

        Returns:
          sequence of Leaf:
        """
        return [entry for entry in self.walk() if isinstance(entry, Leaf)]

    def leaf_count(self):
        """Returns the total number of leaves in this MST.

        Returns:
          int:
        """
        return len(self.leaves())


#     Reachable tree traversal
#     -------------------

    # Walk reachable branches of tree & emit nodes, consumer can bail at any
    # point by returning False

#     def walk_reachable(): AsyncIterable<NodeEntry>:
#         yield self
#         for entry in self.get_entries():
#             if isinstance(entry, MST):
#                 try:
#                     for e in entry.walk_reachable():
#                         yield e
#                 catch (err):
#                     if err instanceof MissingBlockError:
#                         continue
#                     else:
#                         raise err
#             else:
#                 yield entry

#     def reachable_leaves():
#     """
#     Returns:
#       Leaf[]
#     """
#         leaves: Leaf[] = []
#         for entry in self.walk_reachable():
#             if isinstance(entry, Leaf):
#                 leaves.append(entry)
#         return leaves

#     Sync Protocol

    def load_all(self, start=0):
        """Generator. Used in :func:`xrpc_sync.get_repo`.

        (The bluesky-social/atproto TS code calls this ``writeToCarStream``.)

        Args:
          start (int): optional ``subscribeRepos`` sequence number to start from,
            inclusive. Defaults to 0.

        Returns:
          generator of (CID, bytes) tuples
        """
        leaves = set()   # CIDs
        to_fetch = set() # CIDs

        pointer = self.get_pointer()
        assert pointer
        to_fetch.add(pointer)

        while to_fetch:
            blocks = self.storage.read_many(to_fetch)
            to_fetch.clear()

            for cid, block in blocks.items():
                if block.seq < start:
                    continue

                yield cid, block.encoded
                entries = deserialize_node_data(storage=self.storage,
                                                data=Data(**block.decoded))

                for entry in entries:
                    if isinstance(entry, Leaf):
                        leaves.add(entry.value)
                    else:
                        to_fetch.add(entry.get_pointer())

        leaf_blocks = self.storage.read_many(leaves)
        for cid, block in leaf_blocks.items():
            yield cid, block.encoded

#     def cids_for_path(self, key):
#         """Returns the CIDs in a given key path. ???
#
#         Args:
#           key (str):
#
#         Returns:
#           sequence of :class:`CID`
#         """
#         cids: CID[] = [self.get_pointer()]
#         index = self.find_gt_or_equal_leaf_index(key)
#         found = self.at_index(index)
#         if found and isinstance(found, Leaf) and found.key == key:
#             return cids + [found.value]
#         prev = self.at_index(index - 1)
#         if prev and isinstance(prev, MST):
#             return cids + prev.cids_for_path(key)
#         return cids


def leading_zeros_on_hash(key):
    """Returns the number of leading zeros in a key's hash.

    Args:
      key (str or bytes)

    Returns:
      int:
    """
    if not isinstance(key, bytes):
        key = key.encode()  # ensure_valid_key enforces that this is ASCII only

    leading_zeros = 0
    for byte in sha256(key).digest():
        if byte < 64:
             leading_zeros += 1
        if byte < 16:
             leading_zeros += 1
        if byte < 4:
             leading_zeros += 1
        if byte == 0:
            leading_zeros += 1
        else:
            break

    return leading_zeros


def layer_for_entries(entries):
    """
    Args:
      entries (MST or Leaf)

    Returns:
      int or None:
    """
    for entry in entries:
        if isinstance(entry, Leaf):
            return leading_zeros_on_hash(entry.key)


def deserialize_node_data(*, storage=None, data=None, layer=None):
    """
    Args:
      storage (Storage)
      data (Data)

    Returns:
      sequence of MST and Leaf:
    """
    entries = []
    if (data.l is not None):
        entries.append(MST(storage=storage, pointer=data.l,
                           layer=layer - 1 if layer else None))

    last_key = ''
    for entry_data in data.e:
        entry = Entry(**entry_data)
        key_str = entry.k.decode()
        key = last_key[:entry.p] + key_str
        ensure_valid_key(key)
        entries.append(Leaf(key, entry.v))
        last_key = key
        if entry.t is not None:
            entries.append(MST(storage=storage, pointer=entry.t,
                               layer=layer - 1 if layer else None))

    return entries


def serialize_node_data(entries):
    """
    Args:
      entries (sequence of MST and Leaf)

    Returns:
      Data:
    """
    l = None
    i = 0
    if entries and isinstance(entries[0], MST):
        i += 1
        l = entries[0].get_pointer()

    data = Data(l=l, e=[])
    last_key = ''
    while i < len(entries):
        leaf = entries[i]
        next = entries[i + 1] if i < len(entries) - 1 else None

        if not isinstance(leaf, Leaf):
            raise ValueError('Not a valid node: two subtrees next to each other')
        i += 1

        subtree = None
        if next and isinstance(next, MST):
            subtree = next.get_pointer()
            i += 1

        ensure_valid_key(leaf.key)
        prefix_len = common_prefix_len(last_key, leaf.key)
        data.e.append(Entry(
            p=prefix_len,
            k=leaf.key[prefix_len:].encode('ascii'),
            v=leaf.value,
            t=subtree,
        )._asdict())

        last_key = leaf.key

    return data


def common_prefix_len(a, b):
    """
    Args:
      a (str)
      b (str)

    Returns:
      int:
    """
    return len(commonprefix((a, b)))


def cid_for_entries(entries):
    """
    Args:
      entries (sequence of MST and Leaf)

    Returns:
      CID
    """
    return dag_cbor_cid(serialize_node_data(entries)._asdict())


def ensure_valid_key(key):
    """
    Args:
      key (str)

    Raises:
      ValueError: if key is not a valid MST key
    """
    valid = re.compile('[a-zA-Z0-9_\-:.]*$')
    split = key.split('/')
    if not (len(key) <= 256 and
            len(split) == 2 and
            split[0] and
            split[1] and
            valid.match(split[0]) and
            valid.match(split[1])
            ):
        raise ValueError(f'Invalid MST key: {key}')


WalkStatus = namedtuple('WalkStatus', [
    'done',     # bool
    'cur',      # MST or Leaf
    'walking',  # MST or None if cur is the root of the tree
    'index',    # int
], defaults=[None, None, None, None])


class Walker:
    """Allows walking an MST manually.

    Attributes:
      stack (sequence of WalkStatus)
      status (WalkStatus): current
    """
    stack = None
    status = None

    def __init__(self, tree):
        """Constructor.

        Args:
          tree (MST)
        """
        self.stack = []
        self.status = WalkStatus(
            done=False,
            cur=tree,
            walking=None,
            index=0,
        )

    def layer(self):
        """Returns the curent layer of the node we're on."""
        assert not self.status.done, 'Walk is done'

        if self.status.walking:
            return self.status.walking.layer or 0

        # if cur is the root of the tree, add 1
        if isinstance(self.status.cur, MST):
            return (self.status.cur.layer or 0) + 1

        raise RuntimeError('Could not identify layer of walk')


    def step_over(self):
        """Moves to the next node in the subtree, skipping over the subtree."""
        if self.status.done:
            return

        # if stepping over the root of the node, we're done
        if not self.status.walking:
            self.status = WalkStatus(done=True)
            return

        entries = self.status.walking.get_entries()
        self.status = self.status._replace(index=self.status.index + 1)

        if self.status.index >= len(entries):
            if not self.stack:
                self.status = WalkStatus(done=True)
            else:
                self.status = self.stack.pop()
                self.step_over()
        else:
            self.status = self.status._replace(cur=entries[self.status.index])

    def step_into(self):
        """Steps into a subtree.

        Raises:
          RuntimeError: if curently on a leaf
        """
        if self.status.done:
            return

        # edge case for very start of walk
        if not self.status.walking:
            assert isinstance(self.status.cur, MST), \
                'The root of the tree cannot be a leaf'
            next = self.status.cur.at_index(0)
            if not next:
                self.status = WalkStatus(done=True)
            else:
                self.status = WalkStatus(
                    done=False,
                    walking=self.status.cur,
                    cur=next,
                    index=0,
                )
            return

        if not isinstance(self.status.cur, MST):
            raise RuntimeError('No tree at pointer, cannot step into')

        next = self.status.cur.at_index(0)
        assert next, 'Tried to step into a node with 0 entries which is invalid'

        self.stack.append(self.status)
        self.status = WalkStatus(
            walking=self.status.cur,
            cur=next,
            index=0,
            done=False,
        )

    def advance(self):
        """Advances to the next node in the tree.

        Steps into the curent node if necessary.
        """
        if self.status.done:
            return

        if isinstance(self.status.cur, Leaf):
            self.step_over()
        else:
            self.step_into()



"""``com.atproto.repo.*`` XRPC methods."""
import itertools
import json
import logging
import os

import dag_json
from flask import abort, make_response
from lexrpc import Client
from requests import HTTPError

from .repo import Repo, Write
from . import server
from .storage import Action
from . import server
from .util import at_uri, dag_cbor_cid, next_tid, USER_AGENT

logger = logging.getLogger(__name__)


def validate(input, **params):
    input.update(params)

    for field in 'swapCommit', 'swapRecord':
        if input.get(field):
            raise ValueError(f'{field} not supported yet')

    if not input.get('repo'):
        raise ValueError('Missing repo param')


@server.server.method('com.atproto.repo.createRecord')
def create_record(input):
    """Handler for ``com.atproto.repo.createRecord`` XRPC method."""
    validate(input)
    server.auth()

    repo = server.load_repo(input['repo'])
    input.setdefault('rkey', next_tid())
    return put_record(input)


@server.server.method('com.atproto.repo.getRecord')
def get_record(input, repo=None, collection=None, rkey=None, cid=None):
    """Handler for `com.atproto.repo.getRecord` XRPC method."""
    # Largely duplicates xrpc_sync.get_record
    validate(input, repo=repo, collection=collection, rkey=rkey, cid=cid)

    if cid:
        raise ValueError(f'cid not supported yet')

    try:
        repo = server.load_repo(input['repo'])
        record = repo.get_record(collection, rkey)
        if record is not None:
            return json.loads(dag_json.encode({
                'uri': at_uri(repo.did, collection, rkey),
                'cid': dag_cbor_cid(record).encode('base32'),
                'value': record,
            }, dialect='atproto'))
    except ValueError as e:
        logger.info(e)
        pass

    # fall back to AppView if available
    av_host = os.environ.get('APPVIEW_HOST')
    jwt = os.environ.get('APPVIEW_JWT')
    if not av_host or not jwt:
        raise ValueError(f'{collection} {rkey} not found')

    logger.info(f'Falling back to AppView at {av_host}')
    appview = Client(f'https://{av_host}', access_token=jwt,
                     headers={'User-Agent': USER_AGENT})

    try:
        return appview.com.atproto.repo.getRecord(
            {}, repo=input['repo'], collection=collection, rkey=rkey)
    except HTTPError as e:
        body = e.response.json()
        logger.info(f'Returning AppView error to client: {e} {body}')
        status = e.response.status_code
        abort(status, response=make_response(body, status))


@server.server.method('com.atproto.repo.deleteRecord')
def delete_record(input):
    """Handler for ``com.atproto.repo.deleteRecord`` XRPC method."""
    validate(input)
    server.auth()

    repo = server.load_repo(input['repo'])
    record = repo.get_record(input['collection'], input['rkey'])
    if record is None:
        return  # noop

    repo.apply_writes([Write(
        action=Action.DELETE,
        collection=input['collection'],
        rkey=input['rkey'],
    )])


@server.server.method('com.atproto.repo.listRecords')
def list_records(input, repo=None, collection=None, limit=50, cursor=None,
                 reverse=None,
                 # DEPRECATED
                 rkeyStart=None, rkeyEnd=None):
    """Handler for `com.atproto.repo.listRecords` XRPC method.

    KNOWN ISSUE: cursor is interpreted as inclusive, so whenever a cursor is
    used, the response includes the last record returned in the previous
    response.
    """
    validate(input, repo=repo, collection=collection, limit=limit, cursor=cursor)

    if rkeyStart or rkeyEnd:
        raise ValueError(f'rkeyStart/rkeyEnd not supported')
    elif reverse:
        raise ValueError(f'reverse not supported yet')
    elif not collection:
        raise ValueError(f'collection is required')

    repo = server.load_repo(input['repo'])

    start = cursor or f'{collection}/'
    entries = list(itertools.islice(
        itertools.takewhile(lambda entry: entry.key.startswith(f'{collection}/'),
                            repo.mst.walk_leaves_from(key=start)),
        limit))
    blocks = server.storage.read_many([e.value for e in entries])
    records = [blocks[e.value].decoded for e in entries]


    records = [
        json.loads(dag_json.encode({
            'uri': at_uri(repo.did, *entry.key.split('/', 2)),  # collection, rkey
            'cid': dag_cbor_cid(record).encode('base32'),
            'value': record,
        }, dialect='atproto'))
        for entry, record in zip(entries, records)]

    ret = {'records': records}
    if len(entries) == limit:
        ret['cursor'] = entries[-1].key

    return ret


@server.server.method('com.atproto.repo.putRecord')
def put_record(input):
    """Handler for ``com.atproto.repo.putRecord`` XRPC method."""
    validate(input)
    server.auth()

    repo = server.load_repo(input['repo'])
    existing = repo.get_record(input['collection'], input['rkey'])

    repo.apply_writes([Write(
        action=Action.CREATE if existing is None else Action.UPDATE,
        collection=input['collection'],
        rkey=input['rkey'],
        record=input['record'],
    )])

    return {
        'uri': at_uri(repo.did, input['collection'], input['rkey']),
        'cid': dag_cbor_cid(input['record']).encode('base32'),
    }


@server.server.method('com.atproto.repo.describeRepo')
def describe_repo(input, repo=None):
    """Handler for ``com.atproto.repo.describeRepo`` XRPC method."""
    validate(input, repo=repo)
    repo = server.load_repo(input['repo'])

    return {
        'did': repo.did,
        'handle': repo.handle,
        'didDoc': {'TODO': 'TODO'},
        'collections': [
            'app.bsky.actor.profile',
            'TODO',
        ],
        'handleIsCorrect': True,
    }


@server.server.method('com.atproto.repo.applyWrites')
def apply_writes(input):
    """Handler for ``com.atproto.repo.applyWrites`` XRPC method."""
    validate(input)
    server.auth()
    return 'Not implemented yet', 501


@server.server.method('com.atproto.repo.uploadBlob')
def upload_blob(input):
    """Handler for ``com.atproto.repo.uploadBlob`` XRPC method."""
    # input: binary
    validate({})
    server.auth()
    return 'Not implemented yet', 501



"""``com.atproto.server.*`` XRPC methods."""
import logging
import os

from . import server

logger = logging.getLogger(__name__)


@server.server.method('com.atproto.server.createSession')
def create_session(input):
    """Handler for ``com.atproto.server.createSession`` XRPC method."""
    id = input['identifier']
    repo = server.storage.load_repo(id)
    if not repo:
        raise ValueError(f'Repo {id} not found')

    # TODO: generate JWT
    token = os.environ['REPO_TOKEN']
    return {
        'handle': repo.handle,
        'did': repo.did,
        'accessJwt': token,
        'refreshJwt': token,
    }


@server.server.method('com.atproto.server.getSession')
def get_session(input):
    """Handler for ``com.atproto.server.getSession`` XRPC method."""
    server.auth()

    # TODO: parse JWT, extract repo DID
    # decoded = jwt.decode(data, server.repo.privkey, algorithm='ES256K')
    return {
        'handle': server.repo.handle,
        'did': server.repo.did,
    }


@server.server.method('com.atproto.server.refreshSession')
def refresh_session(input, did=None, commit=None):
    """Handler for ``com.atproto.server.refreshSession`` XRPC method."""
    server.auth()

    token = os.environ['REPO_TOKEN']
    return {
        'handle': server.repo.handle,
        'did': server.repo.did,
        'accessJwt': token,
        'refreshJwt': token,
    }


@server.server.method('com.atproto.server.describeServer')
def describe_server(input):
    """Handler for ``com.atproto.server.describeServer`` XRPC method."""
    return {
        'availableUserDomains': [],
        # what is this for?! bsky.social sets it to did:web:bsky.social
        # https://github.com/bluesky-social/atproto/pull/2170#pullrequestreview-1889553896
        'did': f'did:web:{os.environ["PDS_HOST"]}',
    }


@server.server.method('com.atproto.server.getAccountInviteCodes')
def get_account_invite_codes(input, includeUsed=None, createAvailable=None):
    """Handler for ``com.atproto.server.getAccountInviteCodes`` XRPC method."""
    return {'codes': []}


@server.server.method('com.atproto.server.listAppPasswords')
def list_app_passwords(input):
    """Handler for ``com.atproto.server.listAppPasswords`` XRPC method."""
    return {'passwords': []}



"""``com.atproto.sync.*`` XRPC methods."""
from datetime import timedelta, timezone
import itertools
import logging
import os
from threading import Condition
import time

from carbox import car
import dag_cbor
from lexrpc.base import XrpcError
from lexrpc.server import Redirect
from multiformats import CID
from multiformats.multibase import MultibaseKeyError, MultibaseValueError

from .datastore_storage import AtpBlock, AtpRemoteBlob, AtpRepo, DatastoreStorage
from . import server
from .storage import CommitData, SUBSCRIBE_REPOS_NSID
from . import util
from . import xrpc_repo

logger = logging.getLogger(__name__)

# used by subscribe_repos and send_events
NEW_EVENTS_TIMEOUT = timedelta(seconds=60)
new_events = Condition()


@server.server.method('com.atproto.sync.getCheckout')
def get_checkout(input, did=None):
    """Handler for ``com.atproto.sync.getCheckout`` XRPC method.

    Deprecated! Use ``getRepo`` instead.

    Gets a checkout, either head or a specific commit.
    """
    return get_repo(input, did=did)


@server.server.method('com.atproto.sync.getRepo')
def get_repo(input, did=None, since=None):
    """Handler for ``com.atproto.sync.getRepo`` XRPC method."""
    repo = server.load_repo(did)
    start = util.tid_to_int(since) if since else 0

    blocks_and_head = itertools.chain(
        [car.Block(repo.head.cid, repo.head.encoded)],
        (car.Block(cid, data) for cid, data in repo.mst.load_all(start=start)))

    return car.write_car([repo.head.cid], blocks_and_head)


@server.server.method('com.atproto.sync.getRepoStatus')
def get_repo_status(input, did=None):
    """Handler for ``com.atproto.sync.getRepoStatus`` XRPC method."""
    try:
        repo = server.load_repo(did)
    except XrpcError as e:
        if e.name == 'RepoDeactivated':
            return {
                'did': did,
                'active': False,
                'status': 'deactivated',
            }
        raise

    return {
        'did': did,
        'active': True,
    }


@server.server.method('com.atproto.sync.listRepos')
def list_repos(input, limit=500, cursor=None):
    """Handler for ``com.atproto.sync.listRepos`` XRPC method."""
    STATUSES = {'tombstoned': 'deactivated'}

    repos = [{
        'did': repo.did,
        'head': repo.head.cid.encode('base32'),
        'rev': repo.head.seq,
        'active': repo.status is None,
        'status': STATUSES.get(repo.status),
    } for repo in server.storage.load_repos(limit=limit, after=cursor)]

    ret = {'repos': repos}
    if len(repos) == limit:
        ret['cursor'] = repos[-1]['did']

    return ret


def send_events():
    """Triggers ``subscribeRepos`` to deliver new commits from storage to subscribers.
    """
    logger.debug(f'Triggering subscribeRepos to look for new commits')
    with new_events:
        new_events.notify_all()


@server.server.method('com.atproto.sync.subscribeRepos')
def subscribe_repos(cursor=None):
    """Firehose event stream XRPC (ie ``type: subscription``) for all new commits.

    Event stream details: https://atproto.com/specs/event-stream#framing

    This function serves forever, which ties up a runtime context, so it's not
    automatically registered with the XRPC server. Instead, clients should
    choose how to register and serve it themselves, eg asyncio vs threads vs
    WSGI workers.

    See :func:`send_events` for an example thread-based callback to
    register with :class:`Repo` to deliver all new commits to subscribers.
    Here's how to register that callback and this XRPC method in a threaded
    context:

        server.repo.callback = lambda commit_data: xrpc_sync.send_events()
        server.server.register('com.atproto.sync.subscribeRepos',
                               xrpc_sync.subscribe_repos)

    Args:
      cursor (int): try to serve commits from this sequence number forward

    Returns:
      (dict, dict) tuple: (header, payload)
    """
    cur_seq = server.storage.last_seq(SUBSCRIBE_REPOS_NSID)

    def handle(event):
        nonlocal cur_seq

        if isinstance(event, dict):  # non-commit event
            cur_seq = event['seq']
            type = event.pop('$type')
            type_fragment = type.removeprefix('com.atproto.sync.subscribeRepos')
            assert type_fragment != type, type
            return ({'op': 1, 't': type_fragment}, event)

        assert isinstance(event, CommitData), \
            f'unexpected event type {event.__class__} {event}'

        cur_seq = event.commit.seq
        commit = event.commit.decoded
        car_blocks = [car.Block(cid=block.cid, data=block.encoded,
                                decoded=block.decoded)
                      for block in event.blocks.values()]
        return ({  # header
            'op': 1,
            't': '#commit',
        }, {  # payload
            'repo': commit['did'],
            'ops': [{
                'action': op.action.name.lower(),
                'path': op.path,
                'cid': op.cid,
            } for op in (event.commit.ops or [])],
            'commit': event.commit.cid,
            'blocks': car.write_car([event.commit.cid], car_blocks),
            'time': event.commit.time.replace(tzinfo=timezone.utc).isoformat(),
            'seq': event.commit.seq,
            'rev': util.int_to_tid(event.commit.seq, clock_id=0),
            'since': None,  # TODO: load event.commit['prev']'s CID
            'rebase': False,
            'tooBig': False,
            'blobs': [],
        })

    if cursor is not None:
        assert cursor >= 0

    if cursor is not None:
        # validate cursor
        if cursor > cur_seq:
            msg = f'Cursor {cursor} is past our current sequence number {cur_seq}'
            logger.warning(msg)
            yield ({'op': -1}, {'error': 'FutureCursor', 'message': msg})
            return


        if window := os.getenv('ROLLBACK_WINDOW'):
            rollback_start = max(cur_seq - int(window) - 1, 0)
            if cursor < rollback_start:
                logger.warning(f'Cursor {cursor} is before our rollback window; starting at {rollback_start}')
                yield ({'op': 1, 't': '#info'}, {'name': 'OutdatedCursor'})
                cursor = rollback_start

        logger.info(f'fetching existing events from seq {cursor}')
        for event in server.storage.read_events_by_seq(start=cursor):
            yield handle(event)

    # serve new events as they happen. if we see a sequence number skipped, wait
    # for it up to NEW_EVENTS_TIMEOUT before giving up on it and moving on
    logger.info(f'serving new events')
    timeout_s = NEW_EVENTS_TIMEOUT.total_seconds()
    last_yield = time.time()

    while True:
        with new_events:
            new_events.wait(timeout_s)

        for commit_data in server.storage.read_events_by_seq(start=cur_seq + 1):
            last_seq = cur_seq
            event = handle(commit_data)

            waited_enough = time.time() - last_yield > timeout_s
            if cur_seq == last_seq + 1 or waited_enough:
                if cur_seq > last_seq + 1:
                    logger.warning(f'Gave up waiting for seqs {last_seq + 1} to {cur_seq - 1}!')

                last_yield = time.time()
                yield event
            else:
                logger.warning(f'Waiting for seq {last_seq + 1}')
                cur_seq = last_seq
                break

        if delay := os.getenv('SUBSCRIBE_REPOS_BATCH_DELAY'):
            time.sleep(float(delay))



@server.server.method('com.atproto.sync.getBlocks')
def get_blocks(input, did=None, cids=()):
    """Handler for ``com.atproto.sync.getBlocks`` XRPC method."""
    repo = server.load_repo(did)

    try:
        cids = [CID.decode(cid) for cid in cids]
    except (MultibaseKeyError, MultibaseValueError):
        raise XrpcError('Invalid CID', name='BlockNotFound')

    car_blocks = []
    blocks = server.storage.read_many(cids)

    for cid in cids:
        block = blocks[cid]
        if block is None:
            raise XrpcError(f'No block found for CID {cid.encode("base32")}',
                            name='BlockNotFound')
        car_blocks.append(car.Block(cid=block.cid, data=block.encoded))

    return car.write_car([server.storage.head], car_blocks)


@server.server.method('com.atproto.sync.getHead')
def get_head(input, did=None):
    """Handler for ``com.atproto.sync.getHead`` XRPC method.

    Deprecated! Use ``getLatestCommit`` instead.
    """
    repo = server.load_repo(did)
    return {
        'root': repo.head.cid.encode('base32'),
    }


@server.server.method('com.atproto.sync.getLatestCommit')
def get_latest_commit(input, did=None):
    """Handler for ``com.atproto.sync.getLatestCommit`` XRPC method."""
    repo = server.load_repo(did)
    return {
        'cid': repo.head.cid.encode('base32'),
        'rev': repo.head.decoded['rev'],
    }


@server.server.method('com.atproto.sync.getRecord')
def get_record(input, did=None, collection=None, rkey=None, commit=None):
    """Handler for ``com.atproto.sync.getRecord`` XRPC method.

    TODO:

    * implement commit
    * merge with xrpc_repo.get_record?
    """
    if commit:
        raise ValueError('commit not supported yet')

    repo = server.load_repo(did)
    record = repo.get_record(collection, rkey)
    if record is None:
        raise ValueError(f'{collection} {rkey} not found')

    block = car.Block(decoded=record)
    return car.write_car([block.cid], [block])


@server.server.method('com.atproto.sync.getBlob')
def get_blob(input, did=None, cid=None):
    r"""Handler for ``com.atproto.sync.getBlob`` XRPC method.

    Right now only supports redirecting to "remote" blobs based on stored
    :class:`AtpRemoteBlob`\s.
    """
    blob = AtpRemoteBlob.query(AtpRemoteBlob.cid == cid).get()
    if blob:
        raise Redirect(to=blob.key.id(), status=301)

    raise ValueError(f'No blob found for CID {cid}')


# @server.server.method('com.atproto.sync.listBlobs')
# def list_blobs(input, did=None, since=None, limit=500):
#     """Handler for ``com.atproto.sync.listBlobs`` XRPC method.
#
#     TODO. The difficulty with this one is that AtpRemoteBlob is
#     repo-independent. Hrm.
#     """
#     # output: {'cids': [CID, ...]}



"""AT Protocol utility for diffing two MSTs.

Heavily based on:
https://github.com/bluesky/atproto/blob/main/packages/repo/src/mst/diff.ts

Huge thanks to the Bluesky team for working in the public, in open source, and to
Daniel Holmgren and Devin Ivy for this code specifically!
"""
from collections import namedtuple
import logging

from .mst import Leaf, MST, Walker

logger = logging.getLogger(__name__)


def mst_diff(cur, prev=None):
    """Generates a diff between two MSTs.

    Args:
        cur (MST)
        prev (MST): optional

    Returns:
        Diff:
    """
    cur.get_pointer()
    if not prev:
        return null_diff(cur)

    prev.get_pointer()
    diff = Diff()

    left_walker = Walker(prev)
    right_walker = Walker(cur)
    while not left_walker.status.done or not right_walker.status.done:
        # if one walker is finished, continue walking the other & logging all nodes
        if left_walker.status.done and not right_walker.status.done:
            node = right_walker.status.cur
            if isinstance(node, Leaf):
                diff.record_add(node.key, node.value)
            else:
                diff.record_new_cid(node.pointer)
            right_walker.advance()
            continue

        elif not left_walker.status.done and right_walker.status.done:
            node = left_walker.status.cur
            if isinstance(node, Leaf):
                diff.record_delete(node.key, node.value)
            else:
                diff.record_removed_cid(node.pointer)
            left_walker.advance()
            continue

        if left_walker.status.done or right_walker.status.done:
            break

        left = left_walker.status.cur
        right = right_walker.status.cur
        if not left or not right:
            break

        # if both pointers are leaves, record an update & advance both or record
        # the lowest key and advance that pointer
        if isinstance(left, Leaf) and isinstance(right, Leaf):
            if left.key == right.key:
                if left.value != right.value:
                    diff.record_update(left.key, left.value, right.value)
                left_walker.advance()
                right_walker.advance()
            elif left.key < right.key:
                diff.record_delete(left.key, left.value)
                left_walker.advance()
            else:
                diff.record_add(right.key, right.value)
                right_walker.advance()

            continue

        # next, ensure that we're on the same layer
        #
        # if one walker is at a higher layer than the other, we need to do one
        # of two things if the higher walker is pointed at a tree, step into
        # that tree to try to catch up with the lower if the higher walker is
        # pointed at a leaf, then advance the lower walker to try to catch up
        # the higher
        if left_walker.layer() > right_walker.layer():
            if isinstance(left, Leaf):
                if isinstance(right, Leaf):
                    diff.record_add(right.key, right.value)
                else:
                    diff.record_new_cid(right.pointer)

                right_walker.advance()
            else:
                diff.record_removed_cid(left.pointer)
                left_walker.step_into()

            continue

        elif left_walker.layer() < right_walker.layer():
            if isinstance(right, Leaf):
                if isinstance(left, Leaf):
                    diff.record_delete(left.key, left.value)
                else:
                    diff.record_removed_cid(left.pointer)

                left_walker.advance()

            else:
                diff.record_new_cid(right.pointer)
                right_walker.step_into()

            continue

        # if we're on the same level, and both pointers are trees, do a
        # comparison. if they're the same, step over. if they're different, step
        # in to find the subdiff
        if isinstance(left, MST) and isinstance(right, MST):
            if left.pointer == right.pointer:
                left_walker.step_over()
                right_walker.step_over()
            else:
                diff.record_new_cid(right.pointer)
                diff.record_removed_cid(left.pointer)
                left_walker.step_into()
                right_walker.step_into()

            continue

        # finally, if one pointer is a tree and the other is a leaf, simply step
        # into the tree
        if isinstance(left, Leaf) and isinstance(right, MST):
            diff.record_new_cid(right.pointer)
            right_walker.step_into()
            continue

        elif isinstance(left, MST) and isinstance(right, Leaf):
            diff.record_removed_cid(left.pointer)
            left_walker.step_into()
            continue

        raise RuntimeError('Unidentifiable case in diff walk')

    return diff


def null_diff(tree):
    """Generates a "null" diff for a single MST with all adds and new CIDs.

    Args:
        tree (MST)

    Returns:
        Diff:
    """
    diff = Diff()

    for entry in tree.walk():
        if isinstance(entry, Leaf):
            diff.record_add(entry.key, entry.value)
        else:
            diff.record_new_cid(entry.pointer)

    return diff


Change = namedtuple('Change', [
    'key',   # str
    'cid',   # CID
    'prev',  # CID
], defaults=[None])


class Diff:
    """A diff between two MSTs.

    Attributes:
      adds (dict): maps str to :class:`Change`
      updates (dict): maps str to :class:`Change`
      deletes (dict): maps str to :class:`Change`
      new_cids (set of :class:`CID`)
      removed_cids (set of :class:`CID`)
    """

    def __init__(self):
        self.adds = {}
        self.updates = {}
        self.deletes = {}
        self.new_cids = set()
        self.removed_cids = set()

    @staticmethod
    def of(cur, prev=None):
        """
        Args:
          cur (MST)
          prev (MST): optional

        Returns:
          Diff:
        """
        return mst_diff(cur, prev)

    def record_add(self, key, cid):
        """
        Args:
          key (str)
          cid (CID)
        """
        self.adds[key] = Change(key=key, cid=cid)
        self.new_cids.add(cid)

    def record_update(self, key, prev, cid):
        """
        Args:
          key (str)
          prev (CID)
          cid (CID)
        """
        self.updates[key] = Change(key=key, cid=cid, prev=prev)
        self.new_cids.add(cid)

    def record_delete(self, key, cid):
        """
        Args:
          key (str)
          cid (CID)
        """
        self.deletes[key] = Change(key=key, cid=cid)

    def record_new_cid(self, cid):
        """
        Args:
          cid (CID)
        """
        if cid in self.removed_cids:
            self.removed_cids.remove(cid)
        else:
            self.new_cids.add(cid)

    def record_removed_cid(self, cid):
        """
        Args:
          cid (CID)
        """
        if cid in self.new_cids:
            self.new_cids.remove(cid)
        else:
            self.removed_cids.add(cid)

    def add_diff(self, diff):
        """
        Args:
          diff (Diff)
        """
        for add in diff.adds.values():
            if self.deletes[add.key]:
                deleted = self.deletes[add.key]
                if deleted.cid != add.cid:
                    self.record_update(add.key, deleted.cid, add.cid)
                del self.deletes[add.key]
            else:
                self.record_add(add.key, add.cid)

        for update in diff.updates.values():
            self.record_update(update.key, update.prev, update.cid)
            del self.adds[update.key]
            del self.deletes[update.key]

        for deleted in diff.deletes.values():
            if self.adds[deleted.key]:
                del self.adds[deleted.key]
            else:
                del self.updates[deleted.key]
                self.record_delete(deleted.key, deleted.cid)

        self.new_cids |= diff.new_cids

    def updated_keys(self):
        return self.adds | self.updates | self.deletes



"""Google Cloud Datastore implementation of repo storage."""
from datetime import timezone
from functools import wraps
import json
import logging
import mimetypes
import requests

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
import dag_cbor
import dag_json
from google.cloud import ndb
from google.cloud.ndb import context
from google.cloud.ndb.exceptions import ContextError
from lexrpc import ValidationError
from multiformats import CID, multicodec, multihash

from .mst import MST
from .repo import Repo
from .server import server
from . import storage
from .storage import Action, Block, Storage, SUBSCRIBE_REPOS_NSID
from .util import (
    dag_cbor_cid,
    tid_to_int,
    DEACTIVATED,
    DELETED,
    TOMBSTONED,
    TombstonedRepo,
)

logger = logging.getLogger(__name__)


class WriteOnce:
    """:class:`ndb.Property` mix-in, prevents changing it once it's set."""
    def _set_value(self, entity, value):
        existing = self._get_value(entity)
        if existing is not None and value != existing:
            raise ndb.ReadonlyPropertyError(f"{self._name} can't be changed")

        return super()._set_value(entity, value)


class JsonProperty(ndb.TextProperty):
    """Fork of ndb's that subclasses :class:`ndb.TextProperty` instead of :class:`ndb.BlobProperty`.

    This makes values show up as normal, human-readable, serialized JSON in the
    web console.
    https://github.com/googleapis/python-ndb/issues/874#issuecomment-1442753255

    Duplicated in oauth-dropins/webutil:
    https://github.com/snarfed/webutil/blob/main/models.py
    """
    def _validate(self, value):
        if not isinstance(value, dict):
            raise TypeError('JSON property must be a dict')

    def _to_base_type(self, value):
        as_str = json.dumps(value, separators=(',', ':'), ensure_ascii=True)
        return as_str.encode('ascii')

    def _from_base_type(self, value):
        if not isinstance(value, str):
            value = value.decode('ascii')
        return json.loads(value)


class ComputedJsonProperty(JsonProperty, ndb.ComputedProperty):
    """Custom :class:`ndb.ComputedProperty` for JSON values that stores them as
    strings.

    ...instead of like :class:`ndb.StructuredProperty`, with "entity" type, which
    bloats them unnecessarily in the datastore.
    """
    def __init__(self, *args, **kwargs):
        kwargs['indexed'] = False
        super().__init__(*args, **kwargs)


class WriteOnceBlobProperty(WriteOnce, ndb.BlobProperty):
    pass


class CommitOp(ndb.Model):
    """Repo operations - creates, updates, deletes - included in a commit.

    Used in a :class:`StructuredProperty` inside :class:`AtpBlock`; not stored
    directly in the datastore.

    https://googleapis.dev/python/python-ndb/latest/model.html#google.cloud.ndb.model.StructuredProperty
    """
    action = ndb.StringProperty(required=True, choices=('create', 'update', 'delete'))
    path = ndb.StringProperty(required=True)
    cid = ndb.StringProperty()  # unset for deletes


class AtpRepo(ndb.Model):
    r"""An ATProto repo.

    Key name is DID. Only stores the repo's metadata. Blocks are stored in
    :class:`AtpBlock`\s.

    Attributes:
    * handles (str): repeated, optional
    * head (str): CID
    * signing_key (str)
    * rotation_key (str)
    * status (str)
    """
    handles = ndb.StringProperty(repeated=True)
    head = ndb.StringProperty(required=True)
    # TODO: add password hash?

    # these are both secp256k1 private keys, PEM-encoded bytes
    # https://atproto.com/specs/cryptography
    signing_key_pem = ndb.BlobProperty(required=True)
    # TODO: rename this recovery_key_pem?
    # https://discord.com/channels/1097580399187738645/1098725036917002302/1153447354003894372
    rotation_key_pem = ndb.BlobProperty()
    status = ndb.StringProperty(choices=(DEACTIVATED, DELETED, TOMBSTONED))

    created = ndb.DateTimeProperty(auto_now_add=True)
    updated = ndb.DateTimeProperty(auto_now=True)

    @property
    def signing_key(self):
        """(ec.EllipticCurvePrivateKey)"""
        return serialization.load_pem_private_key(self.signing_key_pem,
                                                  password=None)

    @property
    def rotation_key(self):
        """(ec.EllipticCurvePrivateKey` or None)"""
        if self.rotation_key_pem:
            return serialization.load_pem_private_key(self.rotation_key_pem,
                                                      password=None)


class AtpBlock(ndb.Model):
    """A data record, MST node, repo commit, or other event.

    Key name is the DAG-CBOR base32 CID of the data.

    Events should have a fully-qualified ``$type`` field that's one of the
    ``message`` types in ``com.atproto.sync.subscribeRepos``, eg
    ``com.atproto.sync.subscribeRepos#tombstone``.

    Properties:
    * repo (str): DID of the first repo that included this block
    * encoded (bytes): DAG-CBOR encoded value
    * data (dict): DAG-JSON value, only used for human debugging
    * seq (int): sequence number for the subscribeRepos event stream
    """
    repo = ndb.KeyProperty(AtpRepo, required=True)
    encoded = WriteOnceBlobProperty(required=True)
    seq = ndb.IntegerProperty(required=True)
    ops = ndb.StructuredProperty(CommitOp, repeated=True)

    created = ndb.DateTimeProperty(auto_now_add=True)

    @ComputedJsonProperty
    def decoded(self):
        return json.loads(dag_json.encode(dag_cbor.decode(self.encoded)))

    @property
    def cid(self):
        return CID.decode(self.key.id())

    @staticmethod
    def create(*, repo_did, data, seq):
        """Writes a new AtpBlock to the datastore.

        If the block already exists in the datastore, leave it untouched.
        Notably, leave its sequence number as is, since it will be lower than
        this current sequence number.

        Args:
          repo_did (str):
          data (dict): value
          seq (int):

        Returns:
          :class:`AtpBlock`
        """
        assert seq > 0
        encoded = dag_cbor.encode(data)
        digest = multihash.digest(encoded, 'sha2-256')
        cid = CID('base58btc', 1, 'dag-cbor', digest)

        repo_key = ndb.Key(AtpRepo, repo_did)
        atp_block = AtpBlock.get_or_insert(cid.encode('base32'), repo=repo_key,
                                           encoded=encoded, seq=seq)
        assert atp_block.seq <= seq
        return atp_block

    def to_block(self):
        """Converts to :class:`Block`.

        Returns:
          Block
        """
        ops = [storage.CommitOp(action=Action[op.action.upper()], path=op.path,
                                cid=CID.decode(op.cid) if op.cid else None)
               for op in self.ops]
        return Block(cid=self.cid, encoded=self.encoded, seq=self.seq, ops=ops,
                     time=self.created, repo=self.repo)

    @classmethod
    def from_block(cls, *, repo_did, block):
        """Converts a :class:`Block` to an :class:`AtpBlock`.

        Args:
          repo_did (str)
          block (Block)

        Returns:
          AtpBlock
        """
        ops = [CommitOp(action=op.action.name.lower(), path=op.path,
                        cid=op.cid.encode('base32') if op.cid else None)
               for op in (block.ops or [])]
        created = block.time.astimezone(timezone.utc).replace(tzinfo=None)
        return AtpBlock(id=block.cid.encode('base32'), encoded=block.encoded,
                        repo=ndb.Key(AtpRepo, repo_did), seq=block.seq, ops=ops,
                        created=created)


class AtpSequence(ndb.Model):
    """A sequence number for a given event stream NSID.

    Sequence numbers are monotonically increasing, without gaps (which ATProto
    doesn't require), starting at 1. Background:
    https://atproto.com/specs/event-stream#sequence-numbers

    Key name is XRPC method NSID.

    At first, I considered using datastore allocated ids for sequence numbers,
    but they're not guaranteed to be monotonically increasing, so I switched to
    this.
    """
    next = ndb.IntegerProperty(required=True)

    created = ndb.DateTimeProperty(auto_now_add=True)
    updated = ndb.DateTimeProperty(auto_now=True)

    @classmethod
    @ndb.transactional()
    def allocate(cls, nsid):
        """Returns the next sequence number for a given NSID.

        Creates a new :class:`AtpSequence` entity if one doesn't already exist
        for the given NSID.

        Args:
          nsid (str): the subscription XRPC method for this sequence number

        Returns:
          integer, next sequence number for this NSID
        """
        seq = AtpSequence.get_or_insert(nsid, next=1)
        ret = seq.next
        seq.next += 1
        seq.put()
        return ret

    @classmethod
    def last(cls, nsid):
        """Returns the last sequence number for a given NSID.

        Creates a new :class:`AtpSequence` entity if one doesn't already exist
        for the given NSID.

        Args:
          nsid (str): the subscription XRPC method for this sequence number

        Returns:
          integer, last sequence number for this NSID
        """
        seq = AtpSequence.get_or_insert(nsid, next=1)
        return seq.next - 1


class AtpRemoteBlob(ndb.Model):
    """A blob available at a public HTTP URL that we don't store ourselves.

    Key ID is the URL.

    * https://atproto.com/specs/data-model#blob-type
    * https://atproto.com/specs/xrpc#blob-upload-and-download

    TODO:
    * follow redirects, use final URL as key id
    * abstract this in :class:`Storage`
    """
    cid = ndb.StringProperty(required=True)
    size = ndb.IntegerProperty(required=True)
    mime_type = ndb.StringProperty(required=True, default='application/octet-stream')

    created = ndb.DateTimeProperty(auto_now_add=True)
    updated = ndb.DateTimeProperty(auto_now=True)

    @classmethod
    @ndb.transactional()
    def get_or_create(cls, *, url=None, get_fn=requests.get, max_size=None,
                      accept_types=None, name=''):
        """Returns a new or existing :class:`AtpRemoteBlob` for a given URL.

        If there isn't an existing :class:`AtpRemoteBlob`, fetches the URL over
        the network and creates a new one for it.

        Args:
          url (str)
          get_fn (callable): for making HTTP GET requests
          max_size (int, optional): the ``maxSize`` parameter for this blob
            field in its lexicon, if any
          accept_types (sequence of str, optional): the ``accept`` parameter for
            this blob field in its lexicon, if any. The set of allowed MIME types.
          name (str, optional): blob field name in lexicon

        Returns:
          AtpRemoteBlob: existing or newly created blob

        Raises:
          requests.RequestException: if the HTTP request to fetch the blob failed
          lexrpc.ValidationError: if the blob is over ``max_size`` or its type is
            not in ``accept_types``
        """
        def validate_size(size):
            if max_size and size > max_size:
                raise ValidationError(f'{url} Content-Length {size} is over {name} blob maxSize {max_size}')

        assert url
        blob = cls.get_by_id(url)
        if blob:
            validate_size(blob.size)
            server.validate_mime_type(blob.mime_type, accept_types, name=url)
            return blob

        resp = get_fn(url, stream=True)
        resp.raise_for_status()

        mime_type = resp.headers.get('Content-Type')
        if not mime_type:
            mime_type, _ = mimetypes.guess_type(url)
        length = resp.headers.get('Content-Length')
        logger.info(f'Got {resp.status_code} {mime_type} {length} bytes {resp.url}')

        # check type
        server.validate_mime_type(mime_type, accept_types, name=url)

        # check size
        try:
            length = int(length)
        except (TypeError, ValueError):
            length = None  # read body and check length manually below
        if length:
            validate_size(length)

        # now ready to fetch body
        digest = multihash.digest(resp.content, 'sha2-256')
        cid = CID('base58btc', 1, 'raw', digest).encode('base32')

        # note that if the initial URL redirects, we still store it in the
        # AtpRemoteBlob, not the final resolved URL after redirects.
        logger.info(f'Creating new AtpRemoteBlob for {url} CID {cid}')
        blob = cls(id=url, cid=cid, size=len(resp.content))
        if mime_type:
            blob.mime_type = mime_type
        blob.put()

        # re-validate size in case the server didn't give us Content-Length.
        # do this after storing blob so that we don't re-download it next time.
        validate_size(len(resp.content))

        return blob

    def as_object(self):
        """Returns an ATProto ``blob`` object for this blob.

        https://atproto.com/specs/data-model#blob-type

        Returns:
          dict: with ``$type: blob`` and ``ref``, ``mimeType``, and ``size` fields
        """
        return {
            '$type': 'blob',
            'ref': CID.decode(self.cid),
            'mimeType': self.mime_type,
            'size': self.size,
        }


class DatastoreStorage(Storage):
    """Google Cloud Datastore implementation of :class:`Storage`.

    Sequence numbers in :class:`AtpBlock` are allocated per commit; all blocks
    in a given commit will have the same sequence number. They're currently
    sequential counters, starting at 1, stored in an :class:`AtpSequence` entity.

    See :class:`Storage` for method details.
    """
    ndb_client = None

    def __init__(self, *, ndb_client=None):
        """Constructor.

        Args:
          ndb_client (google.cloud.ndb.Client): used in :meth:`read_blocks_by_seq`;
            it's used in the `subscribeRepos` event subscription, so lexrpc calls
            it on a different thread, so it needs its own ndb client context.
        """
        super().__init__()
        self.ndb_client = ndb_client

    def ndb_context(fn):
        @wraps(fn)
        def decorated(self, *args, **kwargs):
            ctx = context.get_context(raise_context_error=False)

            with ctx.use() if ctx else self.ndb_client.context():
                ret = fn(self, *args, **kwargs)

            return ret

        return decorated

    @ndb_context
    def create_repo(self, repo, *, signing_key, rotation_key=None):
        assert repo.did
        assert repo.head

        handles = [repo.handle] if repo.handle else []

        signing_key_pem = signing_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
        rotation_key_pem = None
        if rotation_key:
            rotation_key_pem = rotation_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )

        atp_repo = AtpRepo(id=repo.did, handles=handles,
                           head=repo.head.cid.encode('base32'),
                           signing_key_pem=signing_key_pem,
                           rotation_key_pem=rotation_key_pem)
        atp_repo.put()
        logger.info(f'Stored repo {atp_repo}')

    @ndb_context
    def load_repo(self, did_or_handle):
        assert did_or_handle
        atp_repo = (AtpRepo.get_by_id(did_or_handle)
                    or AtpRepo.query(AtpRepo.handles == did_or_handle).get())

        if not atp_repo:
            logger.info(f"Couldn't find repo for {did_or_handle}")
            return None
        elif atp_repo.status == TOMBSTONED:
            raise TombstonedRepo(f'{atp_repo.key} is tombstoned')

        logger.info(f'Loading repo {atp_repo.key}')
        self.head = CID.decode(atp_repo.head)
        handle = atp_repo.handles[0] if atp_repo.handles else None

        return Repo.load(self, cid=self.head, handle=handle, status=atp_repo.status,
                         signing_key=atp_repo.signing_key,
                         rotation_key=atp_repo.rotation_key)

    @ndb_context
    def load_repos(self, after=None, limit=500):
        query = AtpRepo.query()
        if after:
            query = query.filter(AtpRepo.key > AtpRepo(id=after).key)

        # duplicates parts of Repo.load but batches reading blocks from storage
        atp_repos = query.fetch(limit=limit)

        cids = [CID.decode(r.head) for r in atp_repos]
        blocks = self.read_many(cids)  # dict mapping CID to block
        heads = [blocks[cid] for cid in cids]

        # MST.load doesn't read from storage
        msts = [MST.load(storage=self, cid=block.decoded['data']) for block in heads]
        return [Repo(storage=self, mst=mst, head=head, status=atp_repo.status,
                     handle=atp_repo.handles[0] if atp_repo.handles else None,
                     signing_key=atp_repo.signing_key,
                     rotation_key=atp_repo.rotation_key)
                for atp_repo, head, mst in zip(atp_repos, heads, msts)]

    @ndb_context
    def _set_repo_status(self, repo, status):
        assert status in (DEACTIVATED, DELETED, TOMBSTONED, None)

        @ndb.transactional()
        def update():
            atp_repo = AtpRepo.get_by_id(repo.did)
            atp_repo.status = status
            atp_repo.put()

        update()

    @ndb_context
    def read(self, cid):
        block = AtpBlock.get_by_id(cid.encode('base32'))
        if block:
            return block.to_block()

    @ndb_context
    def read_many(self, cids):
        keys = [ndb.Key(AtpBlock, cid.encode('base32')) for cid in cids]
        got = list(zip(cids, ndb.get_multi(keys)))
        return {cid: block.to_block() if block else None
                for cid, block in got}

    # can't use @ndb_context because this is a generator, not a normal function
    def read_blocks_by_seq(self, start=0, repo=None):
        assert start >= 0

        cur_seq = start
        cur_seq_cids = []

        while True:
            ctx = context.get_context(raise_context_error=False)
            with ctx.use() if ctx else self.ndb_client.context():
                # lexrpc event subscription handlers like subscribeRepos call this
                # on a different thread, so if we're there, we need to create a new
                # ndb context
                try:
                    query = AtpBlock.query(AtpBlock.seq >= cur_seq).order(AtpBlock.seq)
                    if repo:
                        query = query.filter(AtpBlock.repo == AtpRepo(id=repo).key)
                    # unproven hypothesis: need strong consistency to make sure we
                    # get all blocks for a given seq, including commit
                    # https://console.cloud.google.com/errors/detail/CO2g4eLG_tOkZg;service=atproto-hub;time=P1D;refresh=true;locations=global?project=bridgy-federated
                    for atp_block in query.iter(read_consistency=ndb.STRONG):
                        if atp_block.seq != cur_seq:
                            cur_seq = atp_block.seq
                            cur_seq_cids = []
                        if atp_block.key.id() not in cur_seq_cids:
                            cur_seq_cids.append(atp_block.key.id())
                            yield atp_block.to_block()

                    # finished cleanly
                    break

                except ContextError as e:
                    logging.warning(f'lost ndb context! re-querying at {cur_seq}. {e}')
                    # continue loop, restart query

            # Context.use() resets this to the previous context when it exits,
            # but that context is bad now, so make sure we get a new one at the
            # top of the loop
            context._state.context = None

    @ndb_context
    def has(self, cid):
        return self.read(cid) is not None

    @ndb_context
    def write(self, repo_did, obj, seq=None):
        if seq is None:
            seq = self.allocate_seq(SUBSCRIBE_REPOS_NSID)
        return AtpBlock.create(repo_did=repo_did, data=obj, seq=seq).to_block()

    @ndb_context
    @ndb.transactional()
    def apply_commit(self, commit_data):
        seq = tid_to_int(commit_data.commit.decoded['rev'])
        assert seq

        for block in commit_data.blocks.values():
            template = AtpBlock.from_block(
                repo_did=commit_data.commit.decoded['did'], block=block)
            # get_or_insert so we don't wipe out any existing blocks' sequence
            # numbers. (occasionally we see existing blocks recur, eg MST nodes.)
            atp_block = AtpBlock.get_or_insert(
                template.key.id(), repo=template.repo, encoded=block.encoded,
                seq=seq, ops=template.ops)
            block.seq = seq

        self.head = commit_data.commit.cid

        commit = commit_data.commit.decoded
        head_encoded = self.head.encode('base32')

        repo = AtpRepo.get_by_id(commit['did'])
        if repo:
            logger.info(f'Updating {repo.key}')
            repo.head = head_encoded
            repo.put()

    @ndb_context
    def allocate_seq(self, nsid):
        assert nsid
        return AtpSequence.allocate(nsid)

    @ndb_context
    def last_seq(self, nsid):
        assert nsid
        return AtpSequence.last(nsid)



"""Misc AT Protocol utils. TIDs, CIDs, etc."""
import copy
from datetime import datetime, timedelta, timezone
import json
import logging
from numbers import Integral
import random
import re
import time
from urllib.parse import urlparse

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.utils import (
    decode_dss_signature,
    encode_dss_signature,
)
from cryptography.hazmat.primitives import hashes
import dag_cbor
import jwt
from multiformats import CID, multicodec, multihash

logger = logging.getLogger(__name__)

USER_AGENT = 'arroba (https://arroba.readthedocs.io/)'

DOMAIN_RE = re.compile(r'^[^/:;@?!\']+\.[^/:@_?!\']+$')

# the bottom 32 clock ids can be randomized & are not guaranteed to be collision
# resistant. we use the same clockid for all TIDs coming from this runtime.
_clockid = random.randint(0, 31)
_tid_ts_last = 0  # microseconds

S32_CHARS = '234567abcdefghijklmnopqrstuvwxyz'

# for low-S signing
# https://atproto.com/specs/cryptography
CURVE_ORDER = {
    ec.SECP256R1: 0xFFFFFFFF_00000000_FFFFFFFF_FFFFFFFF_BCE6FAAD_A7179E84_F3B9CAC2_FC632551,
    ec.SECP256K1: 0xFFFFFFFF_FFFFFFFF_FFFFFFFF_FFFFFFFE_BAAEDCE6_AF48A03B_BFD25E8C_D0364141
}

DEACTIVATED = 'deactivated'
DELETED = 'deleted'
TOMBSTONED = 'tombstoned'

class TombstonedRepo(ValueError):
    def __init__(self, message='Repo is tombstoned', *args, **kwargs):
        super().__init__(message, *args, **kwargs)


def now(tz=timezone.utc, **kwargs):
    """Wrapper for :meth:`datetime.datetime.now` that lets us mock it out in tests."""
    return datetime.now(tz=tz, **kwargs)


def time_ns():
    """Wrapper for :func:`time.time_ns` that lets us mock it out in tests."""
    return time.time_ns()


def dag_cbor_cid(obj):
    """Returns the DAG-CBOR CID for a given object.

    Args:
      obj: CBOR-compatible native object or value

    Returns:
      CID:
    """
    encoded = dag_cbor.encode(obj)
    digest = multihash.digest(encoded, 'sha2-256')
    return CID('base58btc', 1, 'dag-cbor', digest)


def s32encode(num):
    """Base32 encode with encoding variant sort.

    Based on https://github.com/bluesky-social/atproto/blob/main/packages/common-web/src/tid.ts

    Args:
      num (int or Integral)

    Returns:
      str:
    """
    assert isinstance(num, Integral)

    encoded = []
    while num > 0:
        c = num % 32
        num = num // 32
        encoded.insert(0, S32_CHARS[c])

    return ''.join(encoded)


def s32decode(val):
    """Base32 decode with encoding variant sort.

    Based on https://github.com/bluesky-social/atproto/blob/main/packages/common-web/src/tid.ts

    Args:
      val (str)

    Returns:
      int or Integral:
    """
    i = 0
    for c in val:
        i = i * 32 + S32_CHARS.index(c)

    return i


def datetime_to_tid(dt, clock_id=None):
    """Converts a datetime to a TID.

    https://atproto.com/specs/record-key#record-key-type-tid

    Args:
      dt (datetime.datetime):
      clock_id: 0, optional. If not specified, uses this runtime's clock id

    Returns:
      str: base32-encoded TID
    """
    return int_to_tid(int(dt.timestamp() * 1000 * 1000), clock_id=clock_id)


def int_to_tid(num, clock_id=None):
    """Converts an integer to a TID.

    https://atproto.com/specs/record-key#record-key-type-tid

    Args:
      seq (int)
      clock_id (int): optional. If not specified, uses this runtime's clock id

    Returns:
      str: base32-encoded TID
    """
    if clock_id is None:
        clock_id = _clockid

    tid = (s32encode(num) +
           s32encode(clock_id).ljust(2, '2'))  # clock id
    assert len(tid) <= 13, tid
    return tid.rjust(13, '2')  # pad up to 13 chars; '2' decodes to 0


def tid_to_datetime(tid):
    """Converts an TID to a datetime.

    https://atproto.com/specs/record-key#record-key-type-tid

    Args:
      tid (bytes): base32-encoded TID

    Returns:
      datetime.datetime:

    Raises:
      ValueError: if tid is not bytes or not 13 characters long
    """
    return datetime.fromtimestamp(tid_to_int(tid) / 1000 / 1000, timezone.utc)


def tid_to_int(tid):
    """Converts an TID to an integer.

    https://atproto.com/specs/record-key#record-key-type-tid

    Args:
      tid (bytes): base32-encoded TID

    Returns:
      int:

    Raises:
      ValueError: if tid is not bytes or not 13 characters long
    """
    if not isinstance(tid, (str, bytes)) or len(tid) != 13:
        raise ValueError(f'Expected 13-character str or bytes; got {tid}')

    encoded = tid.replace('-', '')[:-2]  # strip clock id
    return s32decode(encoded)


def next_tid():
    """Returns the TID corresponding to the current time.

    A TID is a base32-sortable-encoded UNIX timestamp (ie time since the epoch)
    in microseconds. Returned tids are guaranteed to monotonically increase
    across calls.

    https://atproto.com/specs/atp#timestamp-ids-tid
    https://atproto.com/specs/record-key#record-key-type-tid
    https://github.com/bluesky-social/atproto/blob/main/packages/common-web/src/tid.ts

    Returns:
      str: TID
    """
    global _tid_ts_last

    # enforce that we're at least 1us after the last TID to prevent TIDs moving
    # backwards if system clock drifts backwards
    _tid_ts_last = max(time_ns() // 1000, _tid_ts_last + 1)
    return int_to_tid(_tid_ts_last)


def at_uri(did, collection, rkey):
    """Returns the ``at://`` URI for a given DID, collection, and rkey.

    https://atproto.com/specs/at-uri-scheme

    Args:
      did (str)
      collection (str)
      rkey (str)

    Returns:
      str: ``at://`` URI
    """
    assert did
    assert collection
    assert rkey
    return f'at://{did}/{collection}/{rkey}'


def parse_at_uri(uri):
    """Parses the repo DID, collection, and rkey out of an at:// URI.

    https://atproto.com/specs/at-uri-scheme

    Args:
      uri (str)

    Returns:
      tuple of str: ``(did, collection, rkey)``
    """
    if not uri or not uri.startswith('at://'):
        raise ValueError(f"{uri} isn't an at:// URI")

    parsed = urlparse(uri)
    rkey = parsed.path.strip('/').split('/', maxsplit=1)
    if len(rkey) < 2:
        rkey.append('')
    return parsed.netloc, *rkey


def new_key(seed=None):
    """Generates a new ECC K-256 keypair.

    https://atproto.com/specs/cryptography

    Args:
      seed (int): optional deterministic value to derive private key from.
        Don't use in production!

    Returns:
      ec.EllipticCurvePrivateKey:
    """
    if seed:
        return ec.derive_private_key(seed, ec.SECP256K1())
    else:
        return ec.generate_private_key(ec.SECP256K1())


def sign(obj, private_key):
    """Signs an object, eg a repo commit or DID document.

    Adds the signature in the ``sig`` field.

    https://atproto.com/specs/cryptography

    The signature is ECDSA around SHA-256 of the input, including a custom
    second pass to enforce that it's the "low-S" variant:
    https://atproto.com/specs/cryptography#ecdsa-signature-malleability

    Args:
      obj (dict)
      private_key (cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey)

    Returns:
      dict: ``obj`` with new ``sig`` field
    """
    orig_sig = private_key.sign(dag_cbor.encode(obj), ec.ECDSA(hashes.SHA256()))
    r, s = decode_dss_signature(apply_low_s_mitigation(orig_sig, private_key.curve))
    obj['sig'] = r.to_bytes(32, 'big') + s.to_bytes(32, 'big')
    return obj

    # old, using pycryptodome
    # signer = DSS.new(private_key, 'fips-186-3', randfunc=_randfunc)
    # commit['sig'] = signer.sign(SHA256.new(dag_cbor.encode(commit)))
    # return commit


def apply_low_s_mitigation(signature, curve):
    """Low-S signature mitigation.

    This prevents signature malleability. (It *doesn't* guarantee deterministic
    signatures though!)

    https://atproto.com/specs/cryptography#ecdsa-signature-malleability

    From picopds. Thank you David!
    https://github.com/DavidBuchanan314/picopds/blob/main/signing.py

    Args:
      signature (bytes)
      curve (ec.EllipticCurve)

    Returns:
      bytes:
    """
    r, s = decode_dss_signature(signature)
    n = CURVE_ORDER[type(curve)]
    if s > n // 2:
        s = n - s
    return encode_dss_signature(r, s)


def verify_sig(obj, public_key):
    """Returns True if obj's signature is valid, False otherwise.

    See :func:`sign` for more background.

    Args:
      obj (dict): repo commit
      public_key (ec.EllipticCurvePublicKey):

    Raises:
      KeyError: if ``obj`` isn't signed, ie doesn't have a ``sig`` field
    """
    obj = copy.copy(obj)
    sig = obj.pop('sig')

    if len(sig) != 64:
        logger.debug(f'Expected signature to be 64 bytes, got {len(sig)}')
        return False

    r = int.from_bytes(sig[:32], 'big')
    s = int.from_bytes(sig[32:], 'big')
    der_sig = encode_dss_signature(r, s)

    try:
        public_key.verify(der_sig, dag_cbor.encode(obj), ec.ECDSA(hashes.SHA256()))
        return True
    except InvalidSignature:
        logger.debug("Couldn't verify signature", exc_info=True)
        return False


def service_jwt(host, repo_did, privkey, expiration=timedelta(minutes=10),
                aud=None, **claims):
    """Generates an inter-service JWT, eg for a relay or AppView.

    https://atproto.com/specs/xrpc#inter-service-authentication-temporary-specification

    Args:
      host (str): hostname of the service this JWT is for, eg ``bgs.bsky-sandbox.dev``
      repo_did (str): DID of the repo this JWT is for
      privkey (ec.EllipticCurvePrivateKey): repo's signing key
      expiration (timedelta): length of time this JWT will be valid, defaults to 10m
      aud (str): JWT audience. Default is ``did:web:[host]``, which works for relays
        and AppViews, but others (eg mod services) have ``did:plc``s instead.
      claims (dict): additional claims to include in the JWT, eg ``lxm``

    Returns:
      str: JWT
    """
    assert host
    assert repo_did
    assert expiration
    data = {
        'iss': repo_did,
        'aud': aud or f'did:web:{host}',
        'alg': 'ES256K',  # k256
        'exp': int((now() + expiration).timestamp()),
        **claims,
    }
    logger.info(f'Generating ATProto inter-service JWT: {data}')
    return jwt.encode(data, privkey, algorithm='ES256K')



"""Utilities to create and resolve did:plcs, did:webs, and handles.

* https://www.w3.org/TR/did-core/
* https://atproto.com/specs/did-plc
* https://github.com/bluesky-social/did-method-plc
* https://w3c-ccg.github.io/did-method-web/
* https://atproto.com/specs/handle#handle-resolution
"""
import base64
from collections import namedtuple
from datetime import timedelta
import json
import logging
import os
import re
import urllib.parse

from cachetools import cached, TTLCache
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.hashes import Hash, SHA256
from cryptography.hazmat.primitives import serialization
from dns.exception import DNSException
from dns.rdatatype import TXT
import dns.resolver
import dag_cbor
from multiformats import multibase, multicodec
import requests

from . import util

DidPlc = namedtuple('DidPlc', [
    'did',           # str
    'signing_key',   # ec.EllipticCurvePrivateKey
    'rotation_key',  # ec.EllipticCurvePrivateKey
    'doc',           # dict, DID document
])

logger = logging.getLogger(__name__)

CACHE_SIZE = 5000
CACHE_TTL = timedelta(hours=6)

# from https://atproto.com/specs/handle#handle-identifier-syntax
HANDLE_RE = re.compile(
    r'^([a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?\.)+[a-zA-Z]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?$')


def resolve(did, **kwargs):
    """Resolves a ``did:plc`` or ``did:web``.

    Args:
      did (str):
      kwargs: passed through to :func:`resolve_plc`/:func:`resolve_web`

    Returns:
      dict: JSON DID document

    Raises:
      ValueError: if the input is not a ``did:plc`` or ``did:web``
      requests.RequestException: if an HTTP request fails
    """
    if did:
        if did.startswith('did:plc:'):
            return resolve_plc(did, **kwargs)
        elif did.startswith('did:web:'):
            return resolve_web(did, **kwargs)

    raise ValueError(f'{did} is not a did:plc or did:web')


@cached(TTLCache(maxsize=CACHE_SIZE, ttl=CACHE_TTL.total_seconds()))
def resolve_plc(did, get_fn=requests.get):
    """Resolves a ``did:plc`` by fetching its DID document from a PLC directory.

    The PLC directory hostname is specified in the ``PLC_HOST`` environment
    variable.

    ``did:plc`` background:

    * https://atproto.com/specs/did-plc
    * https://github.com/bluesky-social/did-method-plc

    Args:
      did (str)
      get_fn (callable): for making HTTP GET requests

    Returns:
      dict: JSON DID document

    Raises:
      ValueError: if the input did is not a ``did:plc`` str
      requests.RequestException: if the HTTP request fails

    """
    if not isinstance(did, str) or not did.startswith('did:plc:'):
        raise ValueError(f'{did} is not a did:plc')

    resp = get_fn(f'https://{os.environ["PLC_HOST"]}/{did}')
    resp.raise_for_status()
    return resp.json()


def create_plc(handle, **kwargs):
    """Creates a new ``did:plc`` in a PLC directory.

    Args are documented in :func:`write_plc`.
    """
    assert 'did' not in kwargs
    assert 'prev' not in kwargs
    return write_plc(handle=handle, **kwargs)


def update_plc(did, **kwargs):
    """Creates a new ``did:plc`` in a PLC directory.

    Args are documented in :func:`write_plc`.
    """
    assert 'rotation_key' in kwargs
    assert 'signing_key' in kwargs

    # get CID of previous head operation for this DID from the directory
    # response is a JSON list with operations from earliest to latest
    # https://github.com/did-method-plc/did-method-plc#audit-logs
    get_fn = kwargs.get('get_fn') or requests.get
    resp = get_fn(f'https://{os.environ["PLC_HOST"]}/{did}/log/audit')
    last_op = resp.json()[-1]

    # merge new data into existing data
    handle = last_op['operation']['alsoKnownAs'].pop(0)
    assert handle.startswith('at://')
    handle = handle.removeprefix('at://')
    kwargs.setdefault('handle', handle)

    kwargs.setdefault('also_known_as', last_op['operation']['alsoKnownAs'])

    # write update operation
    return write_plc(did=did, prev=last_op['cid'], **kwargs)


def write_plc(did=None, handle=None, signing_key=None, rotation_key=None,
              pds_url=None, also_known_as=None, prev=None,
              get_fn=requests.get, post_fn=requests.post):
    """Writes a PLC operation to a PLC directory.

    Generally used to create a new ``did:plc`` or update an existing one.

    The PLC directory hostname is specified in the ``PLC_HOST`` environment
    variable.

    ``did:plc`` background:

    * https://atproto.com/specs/did-plc
    * https://github.com/bluesky-social/did-method-plc

    The DID document in the returned value is the *new format* DID doc, with the
    fully qualified ``verificationMethod.id`` and ``Multikey`` key encoding, ie
    ``did:key`` without the prefix. Details:
    https://github.com/bluesky-social/atproto/discussions/1510

    Args:
      did (str): if provided, updates an existing DID, otherwise creates a new one.
      handle (str): domain handle to associate with this DID
      signing_key (ec.EllipticCurvePrivateKey): The curve must be SECP256K1.
        If omitted, a new keypair will be created.
      rotation_key (ec.EllipticCurvePrivateKey): The curve must be SECP256K1.
        If omitted, a new keypair will be created.
      pds_url (str): PDS base URL to associate with this DID. If omitted,
        defaults to ``https://[PDS_HOST]``
      also_known_as (str or sequence of str): additional URI or URIs to add to
        ``alsoKnownAs``
      prev (str): if an update, the CID of the previous operation for this DID
      get_fn (callable): for making HTTP GET requests
      post_fn (callable): for making HTTP POST requests

    Returns:
      DidPlc: with the newly created ``did:plc``, keys, and DID document

    Raises:
      ValueError: if any inputs are invalid
      requests.RequestException: if the HTTP request to the PLC directory fails
    """
    plc_host = os.environ['PLC_HOST']

    if not isinstance(handle, str) or not handle:
        raise ValueError(f'{handle} is not a valid handle')

    if not pds_url:
        pds_url = f'https://{os.environ["PDS_HOST"]}'

    for key in signing_key, rotation_key:
        if key and not isinstance(key.curve, ec.SECP256K1):
            raise ValueError(f'Expected SECP256K1 key; got {key.curve}')

    if not signing_key:
        logger.info('Generating new k256 signing key')
        signing_key = util.new_key()

    if not rotation_key:
        logger.info('Generating new k256 rotation key')
        rotation_key = util.new_key()

    if not also_known_as:
        also_known_as = []
    elif isinstance(also_known_as, str):
        also_known_as = [also_known_as]

    logger.info('Generating and signing PLC directory operation...')
    # this is a PLC directory genesis operation for creating or updating a DID.
    # it's *not* a DID document. similar but not the same!
    # https://github.com/bluesky-social/did-method-plc#presentation-as-did-document
    op = {
        'type': 'plc_operation',
        'rotationKeys': [encode_did_key(rotation_key.public_key())],
        'verificationMethods': {
            'atproto': encode_did_key(signing_key.public_key()),
        },
        'alsoKnownAs': [f'at://{handle}'] + list(also_known_as),
        'services': {
            'atproto_pds': {
                'type': 'AtprotoPersonalDataServer',
                'endpoint': f'{pds_url}',
            }
        },
        'prev': prev,
    }
    op = util.sign(op, rotation_key)
    op['sig'] = base64.urlsafe_b64encode(op['sig']).decode().rstrip('=')

    if did:
        logger.info(f'Updating existing DID {did}')
    else:
        sha256 = Hash(SHA256())
        sha256.update(dag_cbor.encode(op))
        hash = sha256.finalize()
        did = 'did:plc:' + base64.b32encode(hash)[:24].lower().decode()
        logger.info(f'Creating new DID {did}')

    plc_url = f'https://{plc_host}/{did}'
    logger.info(f'Publishing to {plc_url}  ...')
    resp = post_fn(plc_url, json=op)
    logger.info(f'{resp} {resp.content}')
    resp.raise_for_status()

    op['did'] = did
    return DidPlc(did=did, doc=plc_operation_to_did_doc(op),
                  signing_key=signing_key, rotation_key=rotation_key)


def encode_did_key(pubkey):
    """Encodes an :class:`ec.EllipticCurvePublicKey` into a ``did:key`` string.

    https://atproto.com/specs/did#public-key-encoding

    Args:
      pubkey (ec.EllipticCurvePublicKey)

    Returns:
      str: encoded ``did:key``
    """
    if isinstance(pubkey.curve, ec.SECP256K1):
        codec = 'secp256k1-pub'
    elif isinstance(pubkey.curve, ec.SECP256R1):
        codec = 'p256-pub'
    else:
        raise ValueError(f'Expected secp256k1 or secp256r1 curve, got {pubkey.curve}')

    pubkey_bytes = pubkey.public_bytes(serialization.Encoding.X962,
                                       serialization.PublicFormat.CompressedPoint)
    pubkey_multibase = multibase.encode(multicodec.wrap(codec, pubkey_bytes),
                                        'base58btc')
    return f'did:key:{pubkey_multibase}'


def decode_did_key(did_key):
    """Decodes a ``did:key`` string into an :class:`ec.EllipticCurvePublicKey`.

    https://atproto.com/specs/did#public-key-encoding

    Args:
      did_key (str)

    Returns:
      ec.EllipticCurvePublicKey
    """
    wrapped_bytes = multibase.decode(did_key.removeprefix('did:key:'))
    codec, data = multicodec.unwrap(wrapped_bytes)

    if codec.name == 'secp256k1-pub':
        curve = ec.SECP256K1()
    elif codec.name == 'p256-pub':
        curve = ec.SECP256R1()
    else:
        raise ValueError(f'Expected secp256k1 or secp256r1 curve, got {codec.name}')

    return ec.EllipticCurvePublicKey.from_encoded_point(curve, data)


def plc_operation_to_did_doc(op):
    """Converts a PLC directory operation to a DID document.

    https://github.com/bluesky-social/did-method-plc#presentation-as-did-document

    The DID document in the returned value is the *new format* DID doc, with the
    fully qualified ``verificationMethod.id`` and ``Multikey`` key encoding, ie
    ``did:key`` without the prefix. Details:
    https://github.com/bluesky-social/atproto/discussions/1510

    Args:
      op: dict, PLC operation, https://github.com/did-method-plc/did-method-plc#operation-serialization-signing-and-validation

    Returns:
      dict: DID document, https://www.w3.org/TR/did-core/#data-model
    """
    assert op

    signing_did_key = op['verificationMethods']['atproto']
    return {
        '@context': [
            'https://www.w3.org/ns/did/v1',
            'https://w3id.org/security/multikey/v1',
            'https://w3id.org/security/suites/secp256k1-2019/v1',
        ],
        'id': op['did'],
        'alsoKnownAs': op['alsoKnownAs'],
        'verificationMethod': [{
            'id': f'{op["did"]}#atproto',
            'type': 'EcdsaSecp256r1VerificationKey2019',
            'controller': op['did'],
            'publicKeyMultibase': signing_did_key.removeprefix('did:key:'),
        }],
        'service': [{
            'id': '#atproto_pds',
            'type': 'AtprotoPersonalDataServer',
            'serviceEndpoint': op['services']['atproto_pds']['endpoint'],
        }],
    }


@cached(TTLCache(maxsize=CACHE_SIZE, ttl=CACHE_TTL.total_seconds()))
def resolve_web(did, get_fn=requests.get):
    """Resolves a ``did:web`` by fetching its DID document.

    ``did:web`` spec: https://w3c-ccg.github.io/did-method-web/

    Args:
      did (str)
      get_fn (callable): for making HTTP GET requests

    Returns:
      dict: JSON DID document

    Raises:
      ValueError: if the input did is not a ``did:web`` str
      requests.RequestException: if the HTTP request fails
    """
    if not isinstance(did, str) or not did.startswith('did:web:'):
        raise ValueError(f'{did} is not a did:web')

    did = did.removeprefix('did:web:')
    if ':' in did:
        did = did.replace(':', '/')
    else:
        did += '/.well-known'

    resp = get_fn(f'https://{urllib.parse.unquote(did)}/did.json')
    resp.raise_for_status()
    return resp.json()


@cached(TTLCache(maxsize=CACHE_SIZE, ttl=CACHE_TTL.total_seconds()))
def resolve_handle(handle, get_fn=requests.get):
    """Resolves an ATProto handle to a DID.

    Supports the DNS TXT record and HTTPS well-known methods.

    https://atproto.com/specs/handle#handle-resolution

    Args:
      handle (str)
      get_fn (callable): for making HTTP GET requests

    Returns:
      str or None: DID, or None if the handle can't be resolved

    Raises:
      ValueError: if handle is not a domain
    """
    if not handle or not isinstance(handle, str) or not util.DOMAIN_RE.match(handle):
        raise ValueError(f"{handle} doesn't look like a domain")

    logger.info(f'Resolving handle {handle}')

    # DNS method
    name = f'_atproto.{handle}.'
    try:
        logger.info(f'Querying DNS TXT for {name}')
        answer = dns.resolver.resolve(name, TXT)
        logger.info(f'Got: {answer.response}')
        if answer.canonical_name.to_text() == name:
            for rdata in answer:
                if rdata.rdtype == TXT:
                    text = rdata.to_text()
                    if text.startswith('"did=did:'):
                        return text.strip('"').removeprefix('did=')
    except DNSException as e:
        logger.info(repr(e))

    # HTTPS well-known method
    try:
        resp = get_fn(f'https://{handle}/.well-known/atproto-did')
    except requests.RequestException as e:
        logger.info(f'HTTPS handle resolution failed: {e}')
        return None

    if resp.ok:
        did = resp.text.strip()
        if did.startswith('did:plc:') and len(did.removeprefix('did:plc:')) <= 24:
            return did

    return None



"""Bluesky / AT Protocol repo implementation.

https://atproto.com/guides/data-repos

Heavily based on:
https://github.com/bluesky-social/atproto/blob/main/packages/repo/src/repo.ts

Huge thanks to the Bluesky team for working in the public, in open source, and to
Daniel Holmgren and Devin Ivy for this code specifically!
"""
from collections import defaultdict, namedtuple
import logging

from cryptography.hazmat.primitives.asymmetric import ec
import dag_cbor
from multiformats import CID

from . import util
from .diff import Diff
from .mst import MST
from .server import server
from .storage import (
    Action,
    Block,
    CommitData,
    CommitOp,
    Storage,
    SUBSCRIBE_REPOS_NSID,
)

logger = logging.getLogger(__name__)


Write = namedtuple('Write', [
    'action',      # :class:`Action`
    'collection',  # str
    'rkey',        # str
    'record',      # dict
], defaults=[None] * 4)


def writes_to_commit_ops(writes):
    r"""Converts :class:`Write`\s to :class:`CommitOp`\s.

    Args:
      write (iterable): of :class:`Write`

    Returns:
      list of :class:`repo.CommitOp`
    """
    if not writes:
        return writes

    return [CommitOp(action=write.action,
                     path=f'{write.collection}/{write.rkey}',
                     cid=util.dag_cbor_cid(write.record) if write.record else None)
            for write in writes]


class Repo:
    """AT Protocol data repo implementation, storage agnostic.

    Attributes:
      did (str): repo DID (dynamic property)
      version (int): AT Protocol version (dynamic property)
      storage (Storage)
      mst (MST)
      head (Block): head commit
      handle (str)
      status (str): None (if active), ``'deactivated'``, or ``'tombstoned'``
      callback (callable: (CommitData | dict) => None): called on new commits
        and other repo events. May be set directly by clients. None means no
        callback. The parameter will be a :class:`CommitData` for commits, dict
        record with ``$type`` for other ``com.atproto.sync.subscribeRepos``
        messages.
    """
    storage = None
    mst = None
    head = None
    handle = None
    callback = None
    signing_key = None
    rotation_key = None
    status = None

    def __init__(self, *, storage=None, mst=None, head=None, handle=None,
                 status=None, callback=None, signing_key=None, rotation_key=None):
        """Constructor.

        Args:
          storage (Storage): required
          mst (MST)
          head (dict): head commit
          handle (str)
          status (str): None (if active) or ``'tombstoned'``
          callback (callable, (CommitData | dict) => None)
          signing_key (ec.EllipticCurvePrivateKey): required
          rotation_key (ec.EllipticCurvePrivateKey)
        """
        assert storage
        assert signing_key

        self.storage = storage
        self.mst = mst
        self.head = head
        self.handle = handle
        self.status = status
        self.callback = callback
        self.signing_key = signing_key
        self.rotation_key = rotation_key

    def __eq__(self, other):
        return (self.head and other.head
                and self.version == other.version
                and self.did == other.did
                and self.head == other.head)

    @property
    def did(self):
        if self.head:
            return self.head.decoded['did']

    @property
    def version(self):
        if self.head:
            return self.head.decoded['version']

    def get_record(self, collection, rkey):
        """

        Args:
          collection (str)
          rkey (str)

        Returns:
          dict: node, record or commit or serialized :class:`MST`
        """
        cid = self.mst.get(f'{collection}/{rkey}')
        if cid:
            return self.storage.read(cid).decoded

    def get_contents(self):
        """

        Returns:
          dict mapping str collection to dict mapping str rkey to dict record:
        """
        entries = self.mst.list()
        blocks = self.storage.read_many([e.value for e in entries])
        contents = defaultdict(dict)
        for entry in entries:
            collection, rkey = entry.key.split('/', 2)
            contents[collection][rkey] = blocks[entry.value].decoded

        return contents

    @classmethod
    def create_from_commit(cls, storage, commit_data, *,
                           signing_key, rotation_key=None, **kwargs):
        """

        Args:
          storage (Storage)
          commit_data (CommitData)
          signing_key (ec.EllipticCurvePrivateKey): passed through to
            :meth:`Storage.create_repo`
          rotation_key (ec.EllipticCurvePrivateKey): optional, passed
            through to :meth:`Storage.create_repo`
          kwargs: passed through to :class:`Repo` constructor

        Returns:
          Repo:
        """
        # avoid reading from storage, since if we're in a transaction, those
        # reads won't see writes that happened in apply_commit. instead,
        # construct Repo and MST in memory from existing data.
        mst = MST(storage=storage, pointer=commit_data.commit.decoded['data'])
        repo = Repo(storage=storage, mst=mst, head=commit_data.commit,
                    signing_key=signing_key, rotation_key=rotation_key,
                    **kwargs)

        did = commit_data.commit.repo
        storage.write_event(repo=repo, type='identity', handle=kwargs.get('handle'))
        storage.write_event(repo=repo, type='account', active=True)
        storage.apply_commit(commit_data)

        storage.create_repo(repo, signing_key=signing_key, rotation_key=rotation_key)
        if repo.callback:
            repo.callback(commit_data)

        return repo

    @classmethod
    def create(cls, storage, did, *, signing_key, rotation_key=None,
               initial_writes=None, **kwargs):
        """

        Args:
          storage (Storage)
          did (string)
          signing_key (ec.EllipticCurvePrivateKey): passed through to
            :class:`Storage.create_repo`
          rotation_key (ec.EllipticCurvePrivateKey): optional, passed
            through to :class:`Storage.create_repo`
          initial_writes (sequence of Write)
          kwargs: passed through to :class:`Repo` constructor

        Returns:
          Repo: self
        """
        # initial commit
        commit_data = cls.format_commit(storage=storage, repo_did=did,
                                        signing_key=signing_key,
                                        writes=initial_writes)
        return cls.create_from_commit(storage, commit_data, signing_key=signing_key,
                                      rotation_key=rotation_key, **kwargs)

    @classmethod
    def load(cls, storage, cid=None, **kwargs):
        """
        Args:
          storage (Storage)
          cid (CID): optional
          kwargs: passed through to :class:`Repo` constructor

        Returns:
          Repo:
        """
        commit_cid = cid or storage.head
        assert commit_cid, 'No cid provided and none in storage'

        commit_block = storage.read(commit_cid)
        mst = MST.load(storage=storage, cid=commit_block.decoded['data'])
        logger.info(f'loaded repo for {commit_block.decoded["did"]} at commit {commit_cid}')
        return Repo(storage=storage, mst=mst, head=commit_block, **kwargs)

    @classmethod
    def format_commit(cls, *, repo=None, storage=None, repo_did=None,
                      signing_key=None, mst=None, cur_head=None, writes=None):
        """Creates, but does not store, a new commit.

        If ``repo`` is provided, all other kwargs should be omitted except
        (optionally) ``writes``. Otherwise, ``storage``, ``repo_did``, and
        ``signing_key`` are required.

        If ``repo`` is provided, its ``mst`` attribute is set to the new
        :class:`MST` resulting from applying this commit.

        Args:
          repo (Repo): optional
          storage (Storage): optional
          repo_did (str): optional
          signing_key (ec.EllipticCurvePrivateKey): optional
          mst (MST): optional
          cur_head (CID): optional
          writes (sequence of Write): optional

        Returns:
          CommitData:
        """
        if repo:
            assert (not storage and not repo_did and not signing_key and not mst
                    and not cur_head)
            storage = repo.storage
            repo_did = repo.did
            signing_key = repo.signing_key
            mst = repo.mst
            cur_head = repo.head.cid

        if not mst:
            mst = MST.create(storage=storage)

        commit_blocks = {}  # maps CID to Block
        if writes is None:
            writes = []
        orig_mst = mst

        for write in writes:
            assert isinstance(write, Write), type(write)
            data_key = f'{write.collection}/{write.rkey}'

            if write.action == Action.DELETE:
                mst = mst.delete(data_key)
                continue

            # raises ValidationError if it doesn't validate
            server.validate(write.record.get('$type'), 'record', write.record)

            block = Block(decoded=write.record, repo=repo_did)
            commit_blocks[block.cid] = block
            if write.action == Action.CREATE:
                mst = mst.add(data_key, block.cid)
            else:
                assert write.action == Action.UPDATE
                mst = mst.update(data_key, block.cid)

        root, unstored_blocks = mst.get_unstored_blocks()
        for block in unstored_blocks.values():
            block.repo = repo_did
        commit_blocks.update(unstored_blocks)

        # ensure we're not missing any blocks that were removed and then
        # re-added in this commit
        diff = Diff.of(mst, orig_mst)
        missing = diff.new_cids - commit_blocks.keys()
        if missing:
            commit_blocks.update(storage.read_many(missing))

        commit = util.sign({
            'did': repo_did,
            'version': 3,
            # reuse subscribeRepos sequence number as rev
            # https://github.com/bluesky-social/atproto/discussions/1607
            'rev': util.int_to_tid(storage.allocate_seq(SUBSCRIBE_REPOS_NSID),
                                   clock_id=0),
            'prev': cur_head,
            'data': root,
        }, signing_key)
        commit_block = Block(decoded=commit, ops=writes_to_commit_ops(writes),
                             repo=repo_did)
        commit_blocks[commit_block.cid] = commit_block

        if repo:
            repo.mst = mst

        return CommitData(commit=commit_block, prev=cur_head, blocks=commit_blocks)

    def apply_commit(self, commit_data):
        """

        Args:
          commit_data (CommitData)

        Returns:
          Repo: self
        """
        self.storage.apply_commit(commit_data)
        self.head = commit_data.commit
        if self.callback:
            self.callback(commit_data)
        return self

    def apply_writes(self, writes):
        """

        Args:
          writes (Write or sequence of Write)

        Returns:
          Repo: self
        """
        if isinstance(writes, Write):
            writes = [writes]

        commit_data = Repo.format_commit(repo=self, writes=writes)
        self.apply_commit(commit_data)
        return self
